import logging
import os
import signal
import subprocess
import sys

from pywintypes import error as Win32Error
from win32api import OpenProcess
from win32event import SYNCHRONIZE
from win32file import CloseHandle

from .base import ProgramBase
from chalmers import errors


log = logging.getLogger(__name__)

class NTProgram(ProgramBase):
    """
    Program that implements ProgramBase's abstract
    methods for the win32 platform
    """

    @property
    def is_running(self):

        pid = self.state.get('pid')
        if not pid:
            return False

        try:
            handle = OpenProcess(SYNCHRONIZE, 0, pid)
            CloseHandle(handle)
            return True
        except Win32Error:
            return False


    def start_as_service(self):
        """
        Run this program in a new background process

        windows only
        """

        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE

        from chalmers.scripts import runner as runner_script

        script = os.path.abspath(runner_script.__file__)

        if script.endswith('.pyc') or script.endswith('.pyo'):
            script = script[:-1]

        cmd = [sys.executable, script]

        if 'CHALMERS_ROOT' in os.environ:
            cmd.extend(['--root', os.environ['CHALMERS_ROOT']])

        cmd.append(self.name)
        creationflags = subprocess.CREATE_NEW_CONSOLE | subprocess.CREATE_NEW_PROCESS_GROUP
        p0 = subprocess.Popen(cmd,
                              creationflags=creationflags,
                              startupinfo=startupinfo,
                              close_fds=True)


    def handle_signals(self):
        # Called before keep_alive

        new_mask = self.data.get('umask')
        if new_mask:
            log.warning("Config var 'umask' will be ignored on win32")

        user = self.data.get('user')

        if user:
            raise errors.ChalmersError("Can not yet run as program as a user on win32")


    def dispatch_bg(self):
        raise errors.ChalmersError("Can not yet move a win32 process to the background")

    def _send_signal(self, pid, sig):
        'Kill the process using ctypes and pid'

        import ctypes
        log.error("Send Signal pid=%s sig=%s SIGINT=%s" % (pid, sig, signal.SIGINT))
        if sig == signal.SIGINT:
            os.kill(signal.CTRL_C_EVENT, 0)
            return

        if sig != signal.SIGTERM:
            log.error("Can not kill process with signal %s on windows. Using SIGTERM (%i)" % (sig, signal.SIGTERM))

        PROCESS_TERMINATE = 1
        ExitCode = -1
        handle = ctypes.windll.kernel32.OpenProcess(PROCESS_TERMINATE, False, pid)
        ctypes.windll.kernel32.TerminateProcess(handle, ExitCode)
        ctypes.windll.kernel32.CloseHandle(handle)



"""
Install a crontab rule to run at system boot

E.g.::
    
    @reboot /path/to/python /path/to/chalmers start -a
    
FIXME: There may be cases where the reboot command does not work 
correctly.
 
 * http://unix.stackexchange.com/questions/109804/crontabs-reboot-only-works-for-root
 
"""
from __future__ import unicode_literals, print_function

import logging
import subprocess as sp
import sys

from chalmers import errors
import os
import platform


python_exe = sys.executable
chalmers_script = sys.argv[0]
chalmers_tab_entry = '@reboot %s %s start -a' % (python_exe, chalmers_script)

log = logging.getLogger('chalmers.reboot')

def check():
    try:
        sp.check_output(['crontab', '-l'], stderr=sp.STDOUT)
    except sp.CalledProcessError as err:
        if err.returncode == 1 and 'crontab: no crontab for' in err.output:
            return True
        raise
    except IOError as err:
        if err.errno == 2: return False
        raise
    else:
        return True

def get_crontab():
    try:
        output = sp.check_output(['crontab', '-l']).strip()
    except sp.CalledProcessError as err:
        if err.returncode != 1:
            raise errors.ChalmersError("Could not read crontab")
        return []

    return output.split('\n')

def set_crontab(tab):
    new_cron_tab = '\n'.join(tab) + '\n'

    p0 = sp.Popen(['crontab'], stdin=sp.PIPE)
    p0.communicate(input=new_cron_tab)

py3 = sys.version_info.major == 3
class CronService(object):
    """
    Install a @reboot insruction to the user's local cron table
    """

    def __init__(self, target_user):
        if target_user is not False:
            msg = ("Not implemented: chalmers can not detect "
                   "the init system for your unix machine "
                   "(upstart, systemd or sysv)")
            raise errors.ChalmersError(msg)
        self.target_user = target_user
        log = logging.getLogger('chalmers.cron_service')

        log.info('Platform: %s' % platform.linux_distribution()[0] or 'Unknown')
        log.info('Using posix crond @reboot command')
        log.info('Chalmers service for current user (does not require root)')


    @classmethod
    def use_if_not_root(cls, subcls, target_user):
        if target_user is False:
            if py3:
                return object.__new__(cls)
            else:
                return object.__new__(cls, target_user)

        else:
            if os.getuid() != 0:
                raise errors.ChalmersError("You can not install a posix service for "
                                           "user %s without root privileges. "
                                           "Run this command again with sudo")
            if py3:
                return object.__new__(subcls)
            else:
                return object.__new__(subcls, target_user)

    def install(self):
        tab_lines = get_crontab()

        if chalmers_tab_entry in tab_lines:
            log.warn("Chalmers crontab instruction already exists")
            return True
        else:
            log.info("Adding chalmers instruction to crontab")
            tab_lines.append(chalmers_tab_entry)

            set_crontab(tab_lines)

            log.info("All chalmers programs will now run on boot")
            return True


    def uninstall(self):

        tab_lines = get_crontab()
        if chalmers_tab_entry in tab_lines:
            log.info("Removing chalmers instruction from crontab")
            tab_lines.remove(chalmers_tab_entry)

            set_crontab(tab_lines)
            return True

        else:
            log.info("Chalmers crontab instruction does not exist")
            return False

    def status(self):

        tab_lines = get_crontab()

        if chalmers_tab_entry in tab_lines:
            log.info("Chalmers is setup to start on boot")
            return True
        else:
            log.info("Chalmers will not start on boot")
            return False




'''
Remove a program definition from chalmers
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import sys

from chalmers import errors
from chalmers.utils.cli import add_selection_group, select_programs
from clyent.logs.colors.printer import print_colors


log = logging.getLogger('chalmers.remove')

def main(args):

    programs = select_programs(args, filter_paused=False)

    for prog in programs:
        if not prog.exists():
            print("Program '{0}' does not exist".format(prog.name))
            continue

        print("Removing program {0!s:25} ... ".format(prog.name[:25]), end=''); sys.stdout.flush()

        try:
            prog.delete()
        except errors.ChalmersError as err:
            print_colors("[{=ERROR!c:red} ] %s" % err.message, stream=sys.stdout)
            continue

        print_colors("[  {=OK!c:green}  ]", stream=sys.stdout)

def add_parser(subparsers):

    parser = subparsers.add_parser('remove',
                                      help='Remove a program definition from chalmers',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    add_selection_group(parser)

    parser.set_defaults(main=main)



"""
chalmers.service selects the correct service for the platform on import
"""
from __future__ import absolute_import

import os, platform

if os.name == 'nt':

    from .win32_local_service import Win32LocalService as LocalService
    from .win32_system_service import Win32SystemService as SystemService

    def Service(target_user):
        "Slecet service instance based on target user"
        if target_user is False:
            return LocalService(target_user)
        else:
            return SystemService(target_user)

elif platform.system() == 'Darwin':
    from .darwin_service import DarwinService as SystemService
    LocalService = SystemService
else:
    from .posix_service import PosixSystemService as SystemService, PosixLocalService as LocalService




"""
Chalmers program runner (for windows)

Instead of daemonizing the process like posix,
win32 runs this run script with Popen

Chalmers runs this programm hidden by default
"""

from argparse import ArgumentParser
import logging

from clyent.logs import setup_logging

from chalmers import config
from chalmers.program import Program


logger = logging.getLogger('chalmers')
cli_logger = logging.getLogger('cli-logger')

def main():

    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--root', help='chalmers root config directory')
    parser.add_argument('name', help='name of program to run')
    args = parser.parse_args()

    if args.root:
        config.set_relative_dirs(args.root)


    logfile = config.main_logfile()
    setup_logging(logger, logging.INFO, use_color=False, logfile=logfile, show_tb=True)
    cli_logger.error("Starting program: %s" % args.name)
    prog = Program(args.name)
    prog.start_sync()

if __name__ == '__main__':
    main()



import unittest
from chalmers.utils.kill_tree import kill_tree
from multiprocessing import Process
import time
import os

def child_process():
        p = Process(target=time.sleep, args=(100,))
        p.start()
        p.join()

class Test(unittest.TestCase):

    def test_kills_process(self):

        p = Process(target=time.sleep, args=(100,))
        p.start()
        self.assertTrue(p.is_alive())
        kill_tree(p.pid)
        p.join(1)
        self.assertFalse(p.is_alive())

    def test_does_not_fail(self):

        p = Process(target=child_process)
        p.start()
        self.assertTrue(p.is_alive())
        time.sleep(.1)
        kill_tree(p.pid)
        p.join(1)
        self.assertFalse(p.is_alive())


if __name__ == '__main__':
    unittest.main()




'''
Show the definition file content
'''
from __future__ import unicode_literals, print_function

import logging

from chalmers.program import Program
from chalmers.utils import print_opts
from argparse import RawDescriptionHelpFormatter

log = logging.getLogger('chalmers.show')

def main(args):

    prog = Program(args.name)

    if args.definition:
        print(prog.raw_data.filename)
        return
    if args.state:
        print(prog.state.filename)
        return
    if args.command:
        print(' '.join(prog.raw_data['command']))
        return

    print("Definition file:\n\t%s" % prog.raw_data.filename)
    print("State file:\n\t%s" % prog.state.filename)
    print('State')
    for key, value in prog.state.items():
        print("%12s: %s" % (key, value))

    if args.raw:
        data = prog.raw_data.copy()
    else:
        data = prog.data.copy()

    for category, opts in Program.OPTIONS:
        print_opts(category, data, opts)

    print_opts('Other Options', data, data.keys())


def add_parser(subparsers):
    parser = subparsers.add_parser('show',
                                      help='Show the definition file content',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('name')
    items_group = parser.add_argument_group('Items').add_mutually_exclusive_group()
    items_group.add_argument('--definition', action='store_true',
                        help='Show the definition file name only')
    items_group.add_argument('--state', action='store_true',
                        help='Show the state file name only')
    items_group.add_argument('--command', '--cmd', action='store_true', dest='command',
                        help='Show the command only')
    items_group.add_argument('--status', action='store_true',
                        help='Show the status only')

    parser.add_argument('-r', '--raw', action='store_true',
                        help='Show the definition file before it is formatted and populated with defaults')
    parser.set_defaults(main=main)



"""
Install win32 system service. 

This requires admin privileges to run

 
"""

from __future__ import print_function, absolute_import, unicode_literals

import getpass
import logging

from win32api import GetUserName
from win32com.shell import shell
from win32serviceutil import RemoveService, StopService

from chalmers import errors
from chalmers.event_dispatcher import send_action
from chalmers.program_manager import ProgramManager
from chalmers.windows.install import get_service_name, is_installed, is_running
from chalmers.windows.install import instart


log = logging.getLogger(__name__)

class Win32SystemService(object):
    def __init__(self, target_user):
        if target_user is None:
            target_user = GetUserName()

        self.target_user = target_user

    @property
    def is_admin(self):
        return shell.IsUserAnAdmin()
    @property
    def service_name(self):
        return get_service_name(self.target_user)

    def install(self):
        if not self.is_admin:
            raise errors.ChalmersError("System services requires admin privleges. "
                                       "run this command as an administrator")

        log.info("Your password is required by the windows service manager to launch "
                 "the chalmers service at startup")
        password = getpass.getpass(b"Password for %s: " % self.target_user)

        instart('.\\%s' % self.target_user, password)


    def uninstall(self):
        if not self.is_admin:
            raise errors.ChalmersError("System services requires admin privileges. "
                                       "run this command as an administrator")

        if is_running(self.target_user):
            log.info("Service is running, stopping service %s" % self.service_name)
            StopService(self.service_name)

        if is_installed(self.target_user):
            RemoveService(self.service_name)
            log.info("Uninstalled windows service '%s'" % self.service_name)
        else:
            log.error("Windows service '%s' is not installed" % self.service_name)

    def status(self):

        log.info("Status for service '%s'" % self.service_name)

        if is_installed(self.target_user):
            log.info("service '%s' is installed" % self.service_name)
        else:
            log.error("service '%s' is not installed" % self.service_name)
            return

        if is_running(self.target_user):
            log.info("service '%s' is running" % self.service_name)
        else:
            log.error("service '%s' is not running" % self.service_name)
            return

        pid = send_action(ProgramManager.NAME, "ping")
        log.info("Chalmers manger pid is %s" % pid)





"""
Linux services, this module checks the existence of linux command line 
programs on import

 * systemd_service
 * upstart_service
 * sysv_service
 * cron_service

In that order  
"""
from __future__ import unicode_literals, print_function

import logging
import platform
import sys

from . import cron_service, sysv_service, upstart_service, systemd_service
from chalmers import errors


# Fix for AWS Linux
if sys.version_info.major == 3:
    system_dist = ('system',)
else:
    system_dist = (b'system',)

platform._supported_dists += system_dist


log = logging.getLogger('chalmers.service')

class NoPosixSystemService(object):

    def __init__(self, target_user=None):
        supported_dists = platform._supported_dists + system_dist
        linux = platform.linux_distribution(supported_dists=supported_dists)
        raise errors.ChalmersError("Could not detect system service for platform %s (tried systemd, sysv init and upstart)" % linux[0])

if systemd_service.check():
    PosixSystemService = systemd_service.SystemdService
elif sysv_service.check():
    PosixSystemService = sysv_service.SysVService
elif upstart_service.check():
    PosixSystemService = upstart_service.UpstartService
else:
    PosixSystemService = NoPosixSystemService

PosixLocalService = cron_service.CronService




"""
Install windows startup script

This just writes a script to:: 

    %HOMEPATH%\AppData\Roaming\Microsoft\Windows\Start Menu\Programs\Startup\chalmers.bat

"""
from __future__ import print_function, absolute_import, unicode_literals

import logging
from os import path
import os
import sys

from chalmers import errors
log = logging.getLogger(__name__)

HOME = os.environ.get('HOMEPATH')
STARTUP_DIR = path.join(HOME, 'AppData\Roaming\Microsoft\Windows\Start Menu\Programs\Startup')

class Win32LocalService(object):
    def __init__(self, target_user):

        if target_user is not False:  #
            raise errors.ChalmersError("You can not install a local service into another users account. "
                                       "To do this use the 'runas' windows command with chalmers")

        self.target_user = target_user

    @property
    def script_path(self):
        return path.join(STARTUP_DIR, 'chalmers.bat')

    def install(self):

        chalmers_script = sys.argv[0]
        command = [sys.executable, chalmers_script, 'start', '--all']

        with open(self.script_path, 'w') as fd:
            log.info("Write file: %s" % self.script_path)
            print(' '.join(command), file=fd)

        log.info("All chalmers programs will now run on login")



    def uninstall(self):

        if os.path.exists(self.script_path):
            log.info("Remvoe File: %s" % self.script_path)
            os.unlink(self.script_path)
        else:
            log.warning("File: %s does not exst" % self.script_path)

        log.info("Chalmers local service has been removed")


    def status(self):
        log.info("Check if file exists: %s" % self.script_path)
        if os.path.exists(self.script_path):
            log.info("Chalmers is setup to start on boot")
        else:
            log.info("Chalmers will not start on boot")




import os
from os.path import join
import signal
from subprocess import check_call, Popen
import tempfile
import unittest

import mock

from chalmers import config
from chalmers.program.base import ProgramBase


class TestProgram(ProgramBase):

    def _send_signal(self, pid, signal):
        pass

    @property
    def is_running(self):
        return getattr(self, '_test_is_running', False)

    def start_as_service(self):
        pass

class TestBase(unittest.TestCase):

    def setUp(self):

        self.root_config = join(tempfile.gettempdir(), 'chalmers_tests')
        config.set_relative_dirs(self.root_config)
        unittest.TestCase.setUp(self)

    def test_init(self):

        p = TestProgram('name', load=False)
        self.addCleanup(p.delete)
        self.assertEqual(p.name, 'name')

    def test_create(self):
        p = TestProgram.create('name', {})
        self.addCleanup(p.delete)
        expected_keys = {u'stdout', u'redirect_stderr', u'stopwaitsecs', u'startsecs',
                         u'stopsignal', u'name', u'log_dir', u'startretries',
                         u'daemon_log', u'exitcodes'}

        self.assertEqual(set(p.data.keys()), expected_keys)
        self.assertEqual(p.state, {})

    def test_stopsignal(self):
        p = TestProgram('name')
        self.addCleanup(p.delete)
        self.assertEqual(p.stopsignal, int(signal.SIGTERM))

        p.raw_data.update(stopsignal='foobar')
        p.mk_data()
        self.assertEqual(p.stopsignal, int(signal.SIGTERM))

        p.raw_data.update(stopsignal='SIGINT')
        p.mk_data()
        self.assertEqual(p.stopsignal, int(signal.SIGINT))

        p.raw_data.update(stopsignal=int(signal.SIGINT))
        p.mk_data()
        self.assertEqual(p.stopsignal, int(signal.SIGINT))

    def test_is_ok(self):

        p = TestProgram('name')
        self.addCleanup(p.delete)

        self.assertTrue(p.is_ok)
        p.state.update(exit_status=None)
        self.assertTrue(p.is_ok)

        p.state.update(exit_status=0)
        self.assertTrue(p.is_ok)

        p.state.update(exit_status=1)
        self.assertFalse(p.is_ok)

    def test_text_status(self):

        p = TestProgram('name')
        self.addCleanup(lambda : (setattr(p, '_test_is_running', False), p.delete()))
        self.assertEqual(p.text_status, 'STOPPED')

        p.state.update(paused=True)
        self.assertEqual(p.text_status, 'OFF')

        p.state.update(exit_status=1)
        self.assertEqual(p.text_status, 'ERROR')

        p._test_is_running = True
        self.assertEqual(p.text_status, 'RUNNING')

    def test_setup_output(self):

        p = TestProgram('name')

        if os.path.isfile(p.data['stdout']):
            os.unlink(p.data['stdout'])

        with p.setup_output() as (out, err):
            check_call(['echo', 'hello'], stdout=out, stderr=err)

        with open(p.data['stdout']) as fd:
            output = fd.read()

        self.assertEqual(output.strip(), 'hello')

        with p.setup_output() as (out, err):
            check_call(['echo', 'hello'], stdout=out, stderr=err)

        with open(p.data['stdout']) as fd:
            output = fd.read()

        self.assertEqual(output.split(), ['hello', 'hello'])

    def test_keep_alive_bad_command(self):

        p = TestProgram('name')

        if os.path.isfile(p.data['stdout']):
            os.unlink(p.data['stdout'])

        p.raw_data.update(command=['bad_command'])
        p.mk_data()

        p.keep_alive()

        self.assertEqual(p.state['child_pid'], None)
        self.assertNotEqual(p.state['exit_status'], 0)
        self.assertEqual(p.state['reason'], 'OSError running command "bad_command"')
        self.assertEqual(p.text_status, 'ERROR')


    def test_keep_alive_short(self):

        p = TestProgram('name')

        if os.path.isfile(p.data['stdout']):
            os.unlink(p.data['stdout'])

        p.raw_data.update(command=['echo', 'hi'])
        p.mk_data()

        p.keep_alive()

        self.assertEqual(p.state['child_pid'], None)
        self.assertEqual(p.state['exit_status'], 0)
        self.assertEqual(p.state['reason'], 'Program exited gracefully')
        self.assertEqual(p.text_status, 'STOPPED')

    def test_keep_alive_short_errors(self):

        p = TestProgram('name')

        if os.path.isfile(p.data['stdout']):
            os.unlink(p.data['stdout'])

        p.raw_data.update(command=['echo', 'hi'], exitcodes=[99])
        p.mk_data()

        p.keep_alive()

        self.assertEqual(p.state['child_pid'], None)
        self.assertEqual(p.state['exit_status'], 0)
        self.assertEqual(p.state['reason'], 'Program did not successfully start')
        self.assertEqual(p.text_status, 'ERROR')

    @mock.patch('chalmers.program.base.Popen')
    def test_keep_alive_terminating(self, MockPopen):

        p = TestProgram('name')

        def side_effect(*args, **kwargs):
            p._terminating = True
            return Popen(*args, **kwargs)

        MockPopen.side_effect = side_effect
        if os.path.isfile(p.data['stdout']):
            os.unlink(p.data['stdout'])

        p.raw_data.update(command=['echo', 'hi'])
        p.mk_data()

        p.keep_alive()

        self.assertEqual(p.state['child_pid'], None)
        self.assertEqual(p.state['exit_status'], None)
        self.assertEqual(p.state['reason'], 'Terminated at user request')
        self.assertEqual(p.text_status, 'STOPPED')

    def test_start_sync(self):

        p = TestProgram('name')
        p.keep_alive = lambda: None
        p.start_sync()





if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
[Advanced] [In-development]

import a file exported with `chalmers export`


'''
from __future__ import unicode_literals, print_function

from argparse import FileType
import logging
from os import path
import os

import yaml

from chalmers import config, errors
from chalmers.program import Program


log = logging.getLogger('chalmers.import')


def main(args):

    import_data = yaml.safe_load(args.input)
    groups = {k['group']['name']:k['group'] for k in import_data if 'group' in k}
    programs = {k['program']['name']:k['program'] for k in import_data if 'program' in k}

    for group in groups.values():
        group_dir = path.join(config.dirs.user_data_dir, 'groups')
        if not path.isdir(group_dir): os.makedirs(group_dir)

        group_path = path.join(group_dir, '%s.yaml' % group['name'])
        log.info("Writing group %s to %s" % (group['name'], group_path))
        with open(group_path, 'w') as gf:
            yaml.safe_dump(group, gf, default_flow_style=False)

    for defn in programs.values():
        if 'name' not in defn:
            raise errors.ChalmersError("Import definition requires a name field")

        prog = Program(defn['name'])
        if prog.exists():
            log.warn("Program '%s' already exists, not importing" % defn['name'])
            continue

        prog.raw_data.update(defn)
        prog.mk_data()

def add_parser(subparsers):
    parser = subparsers.add_parser('import',
                                      help='[IN DEVELOPMENT] Batch import many commands from a yaml file',
                                      description=__doc__)

    parser.add_argument('input', type=FileType('r'))

    parser.set_defaults(main=main)



"""

Provide a sudo random default dict of colors 
 
It pics a new color for each lookup that is not already in the colorpicker
"""
from itertools import cycle
from clyent import color

class ColorPicker(object):
    """
    defaultdict like object that populates each new lookup with a color
    """
    light_fg_colors = ['yello', 'white']
    light_bg_colors = [43, 46, 47, 103, 107]
    dark_fg_colors = ['blue', 'red', 'green', 30]
    dark_bg_colors = [40, 41, 42, 44, 45, 100, 102, 104, 105, 106]

    def next_color(self):
        yield color(None, (6,))
        yield color(None, (7,))

        for fg in self.light_fg_colors:
            for bg in self.dark_bg_colors:
                yield color(None, (fg, bg))

        for fg in self.dark_fg_colors:
            for bg in self.light_bg_colors:
                yield color(None, (fg, bg))

    def __init__(self):
        self.color_map = {}
        self.color_iter = iter(cycle(self.next_color()))

    def __getitem__(self, name):
        if name not in self.color_map:
            self.color_map[name] = next(self.color_iter)

        return self.color_map[name]





"""
Install "Systemd" init services definition with systemctl command

http://www.freedesktop.org/wiki/Software/systemd


"""
from __future__ import unicode_literals, print_function

import logging
from os import path
import os
import sys

from chalmers.service.cron_service import CronService
import subprocess as sp
import platform


log = logging.getLogger('chalmers.service')

SYSTEM_D_INIT_DIR = '/etc/systemd/system'


python_exe = sys.executable
chalmers_script = sys.argv[0]

def read_data(filename):
    filename = path.join(path.dirname(__file__), 'data', filename)
    with open(filename) as fd:
        return fd.read()


def check():
    try:
        sp.check_call(['systemctl', '--version'], stdout=sp.PIPE)
        return True

    except OSError as err:
        if err.errno == 2:
            return False
        raise


class SystemdService(object):
    __new__ = CronService.use_if_not_root

    def __init__(self, target_user):
        self.target_user = target_user

        log.info('Platform: %s' % (platform.linux_distribution()[0] or 'Unknown'))
        log.info('Using Linux systemd service')
        if target_user:
            log.info('Chalmers service for target user %s' % target_user)
        else:
            log.info('Chalmers service for root user')


    @property
    def script_name(self):
        if self.target_user:
            return 'chalmers.%s.service' % self.target_user
        else:  # Run as root
            return 'chalmers.service'
    @property
    def script_path(self):
        return path.join(SYSTEM_D_INIT_DIR, self.script_name)

    @property
    def launch_command(self):
        if self.target_user:
            return '/bin/su - %s' % self.target_user
        else:  # Run as root
            return '/bin/sh'

    @property
    def template(self):
        return read_data('systemd.service')

    def check_output(self, command):
        log.info('Running command: %s' % ' '.join(command))
        output = sp.check_output(command)
        log.info(output)
        return output

    def install(self):
        data = self.template.format(python_exe=python_exe,
                                    chalmers=chalmers_script,
                                    launch=self.launch_command)

        with open(self.script_path, 'w') as fd:
            print(data, file=fd)

        log.info("Wrote file: %s" % self.script_path)
        self.check_output(['systemctl', 'enable', self.script_name])

        log.info("Chalmers will now start on boot for user %s" % self.target_user)
        return True

    def uninstall(self):
        try:
            self.check_output(['systemctl', 'disable', self.script_name])
        except sp.CalledProcessError as err:
            if err.returncode == 1:
                return False
            raise

        if path.exists(self.script_path):
            os.unlink(self.script_path)
            log.info("Removed file: %s" % self.script_path)
        else:
            log.warning("Systemd service file %s does not exist" % self.script_path)
        return True

    def status(self):

        try:
            self.check_output(['systemctl', 'is-enabled', self.script_name])
        except sp.CalledProcessError as err:
            if err.returncode == 1:
                log.info("Chalmers will not start on boot")
                return False
            raise
        if not path.exists(self.script_path):
            log.warn("Service file '%s' does not exist " % self.script_path)

        log.info("Chalmers is setup to start on boot for user %s" % self.target_user)
        return True






"""
A program encapsulates all of the actions and functionality need
to operate on and query an underlying process. 
"""

import os

if os.name == 'nt':
    from .nt import NTProgram as Program
else:  # posix
    from .posix import PosixProgram as Program



"""
Event listener based on unix domain sockets or win32 named pipes

"""
from __future__ import print_function

import abc
import logging
from multiprocessing.connection import Listener, Client
import os
import socket
from threading import Thread

from chalmers import errors
from chalmers import config
import sys


log = logging.getLogger(__name__)

try:
    WindowsError
except NameError:
    WindowsError = SystemError

try:
    basestring
except NameError:
    basestring = str

def get_addr(name):
    """
    Get the address to the multiprocessing.connection.Listener or Client objects in
    a platform dependent way
    
    :param name: the name of the chalmers program to connect to
    """

    if os.name == 'nt':
        return r'\\.\pipe\chalmers:%s' % name
    else:
        socdir = os.path.join(config.dirs.user_data_dir, 'sockets')
        if not os.path.isdir(socdir):
            os.makedirs(socdir)

        sock_path = os.path.join(socdir, '%s' % name)

        if sys.version_info.major < 3:
            sock_path = sock_path.encode()

        return sock_path

class EventDispatcher(object):
    """
    This dispatches events listened to from the Listener
    
    """
    __metaclass__ = abc.ABCMeta
    FAMILY = 'AF_PIPE' if os.name == 'nt' else 'AF_UNIX'

    def __init__(self):
        self._running = False
        self._listener = None

    @property
    def listener(self):
        "Return the multiprocessing Listener object"
        if self._listener is None:
            if self.FAMILY == 'AF_UNIX' and os.path.exists(self.addr):
                os.unlink(self.addr)
            try:
                log.debug("Listening to events from: %s" % self.addr)
                self._listener = Listener(self.addr, family=self.FAMILY)
            except socket.error as err:
                if err.errno == 48:
                    msg = "Unix socket '%s' appears to be in use. Please stop this program."
                    raise errors.ChalmersError(msg % self.addr)
                else:
                    raise
        return self._listener

    def start_listener(self):
        "Start listening to the listener in a new thread"
        self._running = True
        self.listener  # Force listener to connect before running in thread
        self._listener_thread = Thread(target=self.listen)
        self._listener_thread.start()

    @abc.abstractproperty
    def name(self):
        return 'chalmers'

    @property
    def addr(self):
        return get_addr(self.name)

    def dispatch_exitloop(self):
        'Exit the listener'
        self._running = False

    def dispatch_ping(self):
        'return the pid of this process'
        return os.getpid()

    @property
    def is_listening(self):
        return self._running

    def listen(self):
        """Listen to incoming clients until
        self._running is set to False
        """
        l = self.listener
        self._running = True
        try:
            while self._running:
                log.debug("Accept connection")
                c = l.accept()

                try:
                    action = c.recv()
                except EOFError:
                    c.close()
                    continue
                if isinstance(action, basestring):
                    args = ()
                    kwargs = {}
                else:
                    args = action.get('args', ())
                    kwargs = action.get('kwargs') or {}
                    action = action.get('action')


                log.info('Dispatch action "%s"' % action)
                method = getattr(self, 'dispatch_%s' % action, None)
                if method:
                    try:
                        result = method(*args, **kwargs)
                    except Exception as err:
                        log.exception(err)
                        c.send({'error':True, 'message':'Exception in action %s - %s' % (action, err)})
                    else:
                        c.send({'error':False, 'message':'ok', 'result': result})
                else:
                    log.warn('No action %s' % action)
                    c.send({'error':True, 'message':'No action %s' % action})
                c.close()
        finally:
            self._listener = None
            l.close()

        log.info("Exiting event loop")

    def send(self, action, *args, **kwargs):
        'Dispatch an action to a Dispatcher'
        return send_action(self.name, action, *args, **kwargs)

def send_action(name, action, *args, **kwargs):
    """
    Send an action to a listener
    the listener must have a 'dispatch_{action}' method
    or this will raise a ChalmersError
    """

    addr = get_addr(name)
    try:
        c = Client(addr, family=EventDispatcher.FAMILY)
    except (socket.error, WindowsError):
        raise errors.ConnectionError("Could not connect to chalmers program %s" % name)

    try:
        c.send({'action': action, 'args':args, 'kwargs':kwargs})
        res = c.recv()

        if res.get('error'):
            raise errors.ChalmersError(res.get('message', 'Unknown error'))
        return res.get('result')
    finally:
        c.close()





"""
Install upstart init services definition with initctl command

http://upstart.ubuntu.com/

"""

import logging
from os import path
import os
from subprocess import check_call, check_output, CalledProcessError, PIPE
import sys

from chalmers.service.cron_service import CronService
import platform


python_exe = sys.executable
chalmers_script = sys.argv[0]

log = logging.getLogger(__name__)

UPSTART_INIT_DIR = '/etc/init'

def read_data(filename):
    filename = path.join(path.dirname(__file__), 'data', filename)
    with open(filename) as fd:
        return fd.read()

def check():
    try:
        check_call(['initctl', '--version'], stdout=PIPE)
        return True
    except OSError as err:
        if err.errno == 2:
            return False
        raise

class UpstartService(object):
    __new__ = CronService.use_if_not_root

    def __init__(self, target_user):
        self.target_user = target_user

        log.info('Platform: %s' % (platform.linux_distribution()[0] or 'Unknown'))
        log.info('Using Linux upstart')
        if target_user:
            log.info('Chalmers service for target user' % target_user)
        else:
            log.info('Chalmers service for root user')

    @property
    def script_name(self):
        if self.target_user:
            return 'chalmers.%s' % self.target_user
        else:  # Run as root
            return 'chalmers'
    @property
    def script_path(self):
        return path.join(UPSTART_INIT_DIR, '%s.conf' % self.script_name)

    @property
    def launch_command(self):
        if self.target_user:
            return '/bin/su - %s' % self.target_user
        else:  # Run as root
            return '/bin/sh'

    def install(self):

        data = read_data('upstart.conf').format(python_exe=python_exe,
                                                chalmers=chalmers_script,
                                                launch=self.launch_command)

        with open(self.script_path, 'w') as fd:
            fd.write(data)

        log.info('Write file: %s' % self.script_path)

    def uninstall(self):
        if path.exists(self.script_path):
            log.info('Remove file: %s' % self.script_path)
            os.unlink(self.script_path)
            log.info("Chalmers service has been removed")
        else:
            log.info("Chalmers service does not exist")

    def status(self):
        #     initctl status chalmers

        command = ['initctl', 'status', self.script_name]
        log.info('Running command: %s' % ' '.join(command))
        try:
            output = check_output(command)
            log.info(output)
        except CalledProcessError as err:
            if err.returncode == 1:
                log.info("Chalmers will not start on boot")
                return False
            raise

        if not path.exists(self.script_path):
            log.warn("Service file '%s' does not exist " % self.script_path)

        log.info("Chalmers is setup to start on boot")

        return True



import abc
import logging
import traceback

import servicemanager
import win32event, win32service, win32api
from win32serviceutil import ServiceFramework


log = logging.getLogger(__name__)

class WindowsService(object, ServiceFramework):
    """
    Base windows service class that provides all the nice things that a python
    service needs
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, args):
        try:

            self._svc_name_ = args[0]
            self._svc_display_name_ = args[0]

            ServiceFramework.__init__(self, args)
            self.stop_event = win32event.CreateEvent(None, 0, 0, None)

        except Exception:
            self.log("Error in WindowsService.__init__")
            self.log(traceback.format_exc())
            raise

    def log(self, msg):
        'Log to the NTEventlog'
        servicemanager.LogInfoMsg(str(msg))

    def sleep(self, sec):
        win32api.Sleep(sec * 1000, True)


    def SvcDoRun(self):
        self.log('start')

        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)
        try:
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
            self.log('start')
            self.start()

            self.ReportServiceStatus(win32service.SERVICE_STOPPED)
#             self.log('wait')
#             win32event.WaitForSingleObject(self.stop_event, win32event.INFINITE)
            self.log('done')
        except Exception:
            self.log("Error in WindowsService.SvcDoRun")
            self.log(traceback.format_exc())
            self.SvcStop()


    def SvcStop(self):
        pass
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        self.log('stopping')
        self.stop()
        self.log('stopped')
        win32event.SetEvent(self.stop_event)
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)





import unittest
from chalmers import utils
import io

class Test(unittest.TestCase):


    def test_try_eval(self):
        self.assertEqual(utils.try_eval('hello'), 'hello')
        self.assertEqual(utils.try_eval('~1"'), '~1"')

    def test_set_nested_key_1(self):
        dct = {}
        utils.set_nested_key(dct, 'key', 'value')
        self.assertEqual(dct, {'key':'value'})

    def test_set_nested_key_2(self):
        dct = {}
        utils.set_nested_key(dct, 'key.key', 'value')
        self.assertEqual(dct, {'key':{'key':'value'}})

    def test_print_opts(self):

        stream = io.StringIO()
        utils.print_opts('category', {'foo':'bar', 'baz':1}, ['foo'], file=stream)

        expected = "\ncategory\n--------\n         foo: 'bar'\n"
        self.assertEqual(stream.getvalue(), expected)

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import os
from os.path import join
import tempfile
import unittest

from chalmers import config

if os.name == 'nt':
    from chalmers.program.nt import NTProgram

@unittest.skipUnless(os.name == 'nt', 'Theses tests only run on win32 systems')
class TestNT(unittest.TestCase):
    
    def setUp(self):

        self.root_config = join(tempfile.gettempdir(), 'chalmers_tests')
        config.set_relative_dirs(self.root_config)
        unittest.TestCase.setUp(self)

    def test_init(self):
        p = NTProgram('foo')
        self.addCleanup(p.delete)

    def test_is_running(self):
        p = NTProgram('foo')
        self.addCleanup(p.delete)

        p.state['pid'] = os.getpid()

        self.assertTrue(p.is_running)

        p.state['pid'] = None

if __name__ == "__main__":
    unittest.main()





"""
Chalmers program object
"""
from __future__ import absolute_import, unicode_literals, print_function

import abc
from contextlib import contextmanager
from glob import glob
import logging
from os import path
import os
import signal
from subprocess import Popen, STDOUT
import sys
from threading import Event
import time

import psutil
import yaml

from chalmers import config
from chalmers import errors
from chalmers.event_dispatcher import EventDispatcher, send_action
from chalmers.utils.file_echo import FileEcho
from chalmers.utils.kill_tree import kill_tree
from chalmers.utils.persistent_dict import PersistentDict


log = logging.getLogger(__name__)

# Python 3
try:
    unicode
except NameError:
    unicode = str

try:
    basestring
except NameError:
    basestring = str


def safe_makedir(filepath):
    dirname = os.path.dirname(filepath)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

def str_replace(data):
    """
    String substitution of `data` dict
    """
    for key, value in data.items():
        if isinstance(value, (str, unicode)):
            data[key] = value.format(**data)

class ProgramBase(EventDispatcher):
    """
    Object that represents a long running process
    
    This program option may represent a program that is 
    running in another process
    or a program that is running in the current process.
    """
    __metaclass__ = abc.ABCMeta


    OPTIONS = [('Primary Options',
                ['name', 'command', 'templates']),
               ('Output',
                ['stdout', 'stderr', 'daemon_log', 'redirect_stderr']),
               ('Process Controll',
                ['retries', 'exitcodes', 'stopwaitsecs',
                 'stopsignal', 'startsecs', 'cwd' ])
              ]
    @property
    def default_data(self):
        return {
                'startretries': 3,
                'exitcodes': [0],
                'startsecs': 3,
                'stopwaitsecs': 10,
                'stopsignal': signal.SIGTERM,
                'log_dir': config.dirs.user_log_dir,

                'redirect_stderr': True,
                'stdout': '{log_dir}/{name}.stdout.log',
                'stderr': '{log_dir}/{name}.stderr.log',
                'daemon_log': '{log_dir}/{name}.daemon.log',
                }

    #===============================================================================
    # Abstract Properties
    #===============================================================================

    @abc.abstractproperty
    def is_running(self): pass

    @abc.abstractmethod
    def _send_signal(self, pid, signal):
        pass

    @abc.abstractmethod
    def start_as_service(self):
        """
        Run this program in a new background process
        
        chalmers manager must be running
        """
        pass

    def handle_signals(self):
        '''
        Can optionally override to handle signals
        '''
        pass

    preexec_fn = None

    def __init__(self, name, load=True, force=False):
        self._name = name

        EventDispatcher.__init__(self)
        self.finished_event = Event()

        defn_filename = path.join(config.dirs.user_data_dir, 'programs', '%s.yaml' % self.name)
        state_filename = path.join(config.dirs.user_data_dir, 'state', '%s.yaml' % self.name)

        try:
            self.state = PersistentDict(state_filename)
        except yaml.error.YAMLError:
            log.error("Yaml parser error. could not parse state file %s" % state_filename)
            if force:
                log.warn("Removing state file and continuing")
                os.unlink(state_filename)
                self.state = PersistentDict(state_filename)
            else:
                msg = "Invalid state file. run `chalmers stop --force` to clear the state file"
                raise errors.ChalmersError(msg)

        try:
            self.raw_data = PersistentDict(defn_filename)
        except yaml.error.YAMLError:
            log.error("Yaml parser error. could not parse definition file %s" % defn_filename)
            if force:
                self.raw_data = PersistentDict(defn_filename, load=False)
            else:
                msg = "Invalid definition file. Run `chalmers edit %s` to fix the definition file"
                raise errors.ChalmersError(msg % (defn_filename))


        self.data = {}
        self.mk_data()

        self._p0 = None
        self.pipe_output = False

    def exists(self):
        return self.raw_data.exists()

    @property
    def name(self):
        return self._name

    @property
    def stopsignal(self):
        stopsignal = self.data['stopsignal']

        if isinstance(stopsignal, basestring) and hasattr(signal, stopsignal):
            stopsignal = getattr(signal, stopsignal)
        if not isinstance(stopsignal, int):
            log.warning("Stopsignal %s is not a valid signal for this platform" % stopsignal)
            log.warning("Signal SIGTERM (%s) will be used" % signal.SIGTERM)
            return signal.SIGTERM
        return stopsignal

    @classmethod
    def load_template(cls, template_name):
        """
        Load a template from file
        """

        template_path = path.join(config.dirs.user_data_dir, 'template', '%s.yaml' % template_name)

        if not path.isfile(template_path):
            return {}

        with open(template_path, 'r') as gf:
            return yaml.safe_load(gf)

    def mk_data(self):
        """
        Transform the 'raw_data' from the definition into
        the used data
        """
        self.data = self.default_data.copy()

        raw_data = self.raw_data or {}

        if 'name' not in raw_data:
            raw_data['name'] = self.name

        for template in raw_data.get('extends', []):
            template_data = self.load_template(template)
            self.data.update(template_data)

        self.data.update(raw_data)

        str_replace(self.data)

        if self.data.get('redirect_stderr'):
            self.data.pop('stderr')

    @property
    def is_paused(self):
        return self.state.get('paused')

    def start(self, daemon=True):
        """
        Start this process 
        :param daemon: if true, start the process as a background process
        
        this will fail if the process is already running
        """
        if self.is_running:
            raise errors.StateError("Process is already running")

        if not daemon:
            self.start_sync()
        else:
            self.start_as_service()

    def start_sync(self):
        """
        Syncronously run this program in this process
        """

        if 'daemon_log' in self.data:
            self.log_to_daemonlog()

        self.start_listener()

        self.state.update(pid=os.getpid())

        try:
            self.keep_alive()
        except errors.StopProcess:
            self._stop()
        finally:
            self.state.update(pid=None, stop_time=time.time())
            self.finished_event.set()
            self._running = False
            if self._listener:
                try:
                    send_action(self.name, 'exitloop')
                except:
                    pass
            if 'daemon_log' in self.data:
                self.stop_daemonlog()


    def dispatch_terminate(self, timeout=None):
        'Action for event listener'
        children = self._stop()
        if not self.finished_event.wait(timeout):
            if self._p0:
                log.info('Process did not stop within %s seconds' % (timeout))
                log.info('Hard killing process %s' % (self._p0.pid))
                kill_tree(self._p0.pid)

            if not self.finished_event.wait(timeout):
                raise errors.ChalmersError("Timed out waiting for program %s to finish" % self.name)
        else:
            for child in children:
                if child.is_running():
                    child.kill()

    def _stop(self):
        """
        Terminate this process, 
        This function may only be called by the process that called 'start_sync'
        
         
        """
        self._running = False
        self._terminating = True

        if not self._p0:
            log.info("_p0 is none")

        try:
            parent = psutil.Process(self._p0.pid)
        except psutil.NoSuchProcess:
            log.info("psutil.NoSuchProcess %s" % self._p0.pid)
            return

        children = parent.children(recursive=True)

        log.info('Sending signal %s to process %s' % (self.stopsignal, self._p0.pid))
        self._send_signal(self._p0.pid, self.stopsignal)

        return children


    @contextmanager
    def setup_output(self):
        if self.pipe_output:
            self.data['redirect_stderr'] = True
        if self.data['redirect_stderr']:
            stderr = STDOUT
        elif self.data.get('stderr'):
            safe_makedir(self.data['stderr'])
            stderr = open(self.data['stderr'], 'a+', 1)
            stderr.seek(0, os.SEEK_END)
        else:
            stderr = None

        if self.data.get('stdout'):
            safe_makedir(self.data['stdout'])
            stdout = open(self.data['stdout'], 'a+', 1)
            stdout.seek(0, os.SEEK_END)
        else:
            stdout = None

        if self.pipe_output:  # TODO: this may no longer be useful
            self._echo = FileEcho(self.data['stdout'], sys.stdout)
            self._echo.start()
        else:
            self._echo = None

        try:
            yield stdout, stderr
        finally:
            if self._echo:
                self._echo.stop()

            if hasattr(stdout, 'close'):
                stdout.close()

            if hasattr(stderr, 'close'):
                stdout.close()

    def keep_alive(self):
        """
        """
        self._p0 = False

        self.handle_signals()

        with self.setup_output() as (stdout, stderr):

            self._terminating = False
            startretries = self.data.get('startretries', 3)
            initial_startretries = self.data.get('startretries', 3)

            while startretries:
                start = time.time()
                if startretries != initial_startretries:
                    log.info('Retry command (%i retries remain)' % startretries)
                env = os.environ.copy()
                update_env = {k:str(v) for (k, v) in self.data.get('env', {}).items()}
                env.update(update_env)
                cwd = self.data.get('cwd') or os.path.abspath(os.curdir)

                env_str = '\n'.join('\t%s: %r' % item for item in update_env.items())

                if env_str:
                    log.info("Setting Environment: \n%s" % env_str)

                if cwd:
                    log.info("Setting Working Directory: %s" % cwd)

                log.info("Running Command: %s" % ' '.join(self.data['command']))
                try:
                    self._p0 = Popen(self.data['command'],
                                     stdout=stdout, stderr=stderr,
                                     env=env, cwd=cwd, bufsize=self.data.get('bufsize', 0),
                                     preexec_fn=self.preexec_fn)
                except OSError as err:
                    log.exception('Program %s could not be started with popen' % self.name)
                    self.state.update(child_pid=None, exit_status=1,
                                      reason='OSError running command "%s"' % self.data['command'][0])
                    return
                except:
                    log.exception('Exception in keep_alive')
                    self.steate.update(child_pid=None, exit_status=1,
                                      reason='There was an unknown exception opening command (check logs)')
                    return

                log.info('Program started with pid %s' % self._p0.pid)
                self.state.update(child_pid=self._p0.pid, reason=None, exit_status=None,
                                  start_time=time.time())

                try:
                    status = self._p0.wait()
                except KeyboardInterrupt:
                    log.error('Program %s was interrupted by user' % self.name)
                    kill_tree(self._p0.pid)
                    self.state.update(child_pid=None, exit_status=None, reason='Interrupted by user')
                    raise
                except BaseException as err:
                    log.error('Program %s was interrupted' % self.name)
                    kill_tree(self._p0.pid)
                    self.state.update(child_pid=None, exit_status=None, reason='Python BaseException')
                    log.exception(err)
                    raise

                self._p0 = False

                uptime = time.time() - start

                log.info('Command Exited with status %s' % status)
                log.info(' + Uptime %s' % uptime)

                if self._terminating:
                    reason = "Terminated at user request"
                    status = None
                elif status in self.data['exitcodes']:
                    reason = "Program exited gracefully"
                elif uptime < self.data['startsecs']:
                    reason = 'Program did not successfully start'
                    startretries -= 1
                else:
                    reason = "Program exited unexpectedly with code %s" % (status)
                    startretries = initial_startretries

                self.state.update(child_pid=None, exit_status=status,
                                  reason=reason)

                if self._terminating:
                    break
                if status in self.data['exitcodes']:
                    break

        log.debug("Exiting keep alive function")

    def delete(self):
        """
        Remove this program definition
        """
        if self.is_running:
            raise errors.ChalmersError("Can not remove running program (must be stopped)")

        self.state.delete()
        self.raw_data.delete()

    @property
    def is_ok(self):
        if self.is_running:
            return True
        elif self.state.get('exit_status') is None:
            return True
        elif self.state.get('exit_status') in self.data['exitcodes']:
            return True
        return False


    @property
    def text_status(self):
        'A text status of the current program'

        if self.is_running:
            return 'RUNNING'
        elif self.is_ok:
            if self.is_paused:
                return 'OFF'
            else:
                return 'STOPPED'
        else:
            return 'ERROR'

    def log_to_daemonlog(self):
        logger = logging.getLogger('chalmers')
        safe_makedir(self.data['daemon_log'])
        self._log_stream = open(self.data['daemon_log'], 'a', 1)
        self._log_stream.seek(0, 2)
        self._daemonlog_hdlr = hdlr = logging.StreamHandler(self._log_stream)
        hdlr.setLevel(logging.INFO)
        fmt = logging.Formatter("[%(asctime)s] %(levelname)s:%(message)s")
        hdlr.setFormatter(fmt)
        logger.setLevel(logging.INFO)
        logger.addHandler(hdlr)

    def stop_daemonlog(self):

        logger = logging.getLogger('chalmers')
        logger.removeHandler(self._daemonlog_hdlr)
        try:
            self._log_stream.close()
        except IOError:
            # Ignore:
            # close() called during concurrent operation on the same file object.
            pass



    def stop(self, force=False):
        """
        Stop this program
        """

        if not self.is_running:
            raise errors.StateError("Program is not running")

        if force:
            if self.state.get('pid'):
                log.debug("Force killing chalmers monitor process pid=%s" % self.state.get('pid'))
                try:
                    kill_tree(self.state.get('pid'))
                except psutil.NoSuchProcess:
                    pass
            if self.state.get('child_pid'):
                log.debug("Force killing chalmers child process pid=%s" % self.state.get('child_pid'))
                try:
                    kill_tree(self.state.get('child_pid'))
                except psutil.NoSuchProcess:
                    pass
        else:
            send_action(self.name, 'terminate', timeout=self.data['stopwaitsecs'])


    def wait_for_start(self):
        """
        Wait for program to start, 
        returns True if the program started successfully
        """
        time.sleep(.5)
        self.state.reload()
        startsecs = self.data['startsecs']
        st = time.time()
        while time.time() - self.state.get('start_time', st) < startsecs:
            time.sleep(1)
            self.state.reload()

        return not self.is_ok

    #===============================================================================
    # Class methods
    #===============================================================================

    @classmethod
    def create(cls, name, defn, state=None):
        """
        Create a new program object
        """
        prog = cls(name, False)
        prog.raw_data.update(defn)
        prog.state.update(state or {})
        prog.mk_data()

        return prog

    @classmethod
    def find_for_user(cls, force=False):
        'Find all programs this user has defined'
        program_glob = path.join(config.dirs.user_data_dir, 'programs', '*.yaml')
        for filename in glob(program_glob):
            basename = path.basename(filename)
            name = path.splitext(basename)[0]
            yield cls(name, force=force)





'''
Add a program without running it

eg:
    chalmers add --name server1 -- python /path/to/myserver.py
or:
    chalmers add --name server1 -c "python /path/to/myserver.py"

Note that this does not run the program by default. To run your program,

run `chalmers start NAME` or use the run-now option eg. `chalmers add --run-now ...`
 
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import os
import shlex

from chalmers import errors
from chalmers.program import Program

log = logging.getLogger('chalmers.add')

def create_definition(args):

    env = {}
    for env_var in args.save_env:
        if env_var in os.environ:
            env[env_var] = os.environ[env_var]
        else:
            log.warn("Environment variable %s does not exist (from -e/--save-env)" % env_var)

    definition = {
                    'name': args.name,
                    'command': args.command,
                    'cwd': os.path.abspath(args.cwd),
                    'env': env,
    }

    if args.stdout:
        definition['stdout'] = args.stdout

    if args.daemon_log:
        definition['daemon_log'] = args.daemon_log

    if args.redirect_stderr is not None:
        definition['redirect_stderr'] = args.redirect_stderr

    if args.stderr is not None:
        definition['stderr'] = args.stderr

    return definition


def main(args):
    if args.cmd and args.command:
        raise errors.ChalmersError('Unknow arguments %r' % args.command)
    elif not (args.cmd or args.command):
        raise errors.ChalmersError('Must specify a command to add')
    if args.cmd:
        args.command = args.cmd

    if not args.name:
        args.name = args.command[0]

    program = Program(args.name)

    if program.exists():
        raise errors.ChalmersError("Program with name '{args.name}' already exists.  \n"
                                   "Use the -n/--name option to change the name or \n"
                                   "Run 'chalmers remove {args.name}' to remove it \n"
                                   "or 'chalmers set' to update the parameters".format(args=args))

    state = {'paused': args.paused}
    definition = create_definition(args)

    program.raw_data.update(definition)
    program.state.update(state)

    if args.run_now:
        program.start()

    log.info('Added program {args.name}'.format(args=args))

def add_parser(subparsers):
    description = 'Add a command to run'
    parser = subparsers.add_parser('add',
                                   help=description, description=description,
                                   epilog=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter
                                      )
    #===============================================================================
    #
    #===============================================================================
    group = parser.add_argument_group('Starting State') \
                  .add_mutually_exclusive_group()

    group.add_argument('--off', '--paused', action='store_true', dest='paused',
                       help="Don't start program automatically at system start (exclude from `chalmers start --all`)",
                       default=False)
    group.add_argument('--on', '--un-paused', action='store_false', dest='paused',
                       help="Start program automatically at system start (include in `chalmers start --all`)")

    group.add_argument('-r', '--run-now', action='store_true', default=False, dest='run_now',
                       help="Start program Right now (default: %(default)s)")
    group.add_argument('-l', '--dont-run-now', '--run-later', action='store_false', dest='run_now',
                       help="Start the program later with `chalmers start ...`")

    #===========================================================================
    #
    #===========================================================================
    group = parser.add_argument_group('Process Output:')
    group.add_argument('--stdout',
                       help='Filename to log stdout to')
    group.add_argument('--stderr',
                       help='Filename to log stderr to')
    group.add_argument('--daemon-log',
                       help='Filename to log meta information about this process to')
    group.add_argument('--redirect-stderr', action='store_true', default=True,
                       dest='redirect_stderr',
                       help='Store stdout and stderr in the same log file (default: %(default)s)')
    group.add_argument('--dont-redirect-stderr', action='store_false',
                       dest='redirect_stderr',
                       help='Store stdout and stderr in seporate log files')
    #===========================================================================
    #
    #===========================================================================
    parser.add_argument('-n', '--name',
                        help='Set the name of this program for future chalmers commands')

    parser.add_argument('--cwd', default=os.curdir,
                        help='Set working directory of the program (default: %(default)s)')

    parser.add_argument('command', nargs='*', metavar='COMMAND',
                        help='Command to run')

    split = lambda item: shlex.split(item, posix=os.name == 'posix')

    parser.add_argument('-c', metavar='COMMAND', type=split, dest='cmd',
                        help='Command to run')

    parser.add_argument('-e', '--save-env', metavar='ENV_VAR', action='append', default=[],
                        help='Save a current environment variable to be run( Eg. --save-env PATH)')

    parser.set_defaults(main=main, state='pause')



from contextlib import contextmanager
import tempfile
import unittest

from chalmers.service.systemd_service import SystemdService, check
from chalmers.service import systemd_service
import subprocess as sp
from os import path

enabled = {}

def mock_check_output(command, *args, **kwargs):
    if command[:2] == ['systemctl', 'is-enabled']:
        if command[2] not in enabled:
            raise sp.CalledProcessError(1, 'systemctl', 'not found')
        return 'ok'
    elif command[:2] == ['systemctl', 'enable']:

        configfile = path.join(systemd_service.SYSTEM_D_INIT_DIR, command[2])
        if not path.exists(configfile):
            raise sp.CalledProcessError(1, 'systemctl', 'not found')
        enabled[command[2]] = True

    elif command[:2] == ['systemctl', 'disable']:

        configfile = path.join(systemd_service.SYSTEM_D_INIT_DIR, command[2])
        if not path.exists(configfile):
            raise sp.CalledProcessError(1, 'systemctl', 'not found')

        del enabled[command[2]]
        return 'ok'
    else:
        raise Exception('unknown command %r' % command)

class Test(unittest.TestCase):

    @contextmanager
    def mock_systemd(self, exists=True):
        from mock import patch
        patch1 = patch('subprocess.check_output')
        patch2 = patch('os.getuid')


        with patch1 as check_output, patch2 as getuid:
            SYSTEM_D_INIT_DIR = systemd_service.SYSTEM_D_INIT_DIR
            systemd_service.SYSTEM_D_INIT_DIR = tempfile.mkdtemp()

            try:
                getuid.return_value = 0

                if  exists:
                    check_output.side_effect = mock_check_output
                else:
                    check_output.side_effect = IOError(2, "err")
                yield
            finally:
                systemd_service.SYSTEM_D_INIT_DIR = SYSTEM_D_INIT_DIR


    def test_install(self):

        with self.mock_systemd(exists=True):

            service = SystemdService(None)
            self.assertFalse(service.status())
            self.assertTrue(service.install())
            self.assertTrue(service.status())
            self.assertTrue(service.uninstall())
            self.assertFalse(service.status())
            self.assertFalse(service.uninstall())



if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_status']
    unittest.main()



"""
Windows win32service class overloaded to run chalmers processes

This service will exit almost immediatly, all processes are
'daemonized' (win32 compatible) and the service will exit while
the processes are still running
"""
from chalmers.program import Program
from chalmers.windows.service_base import WindowsService


class ChalmersService(WindowsService):
    """
    Run the chalmers manager process as a windows service
    """
    def start(self):

        programs = list(Program.find_for_user())
        programs = [p for p in programs if not p.is_paused]

        for prog in programs:
            if not prog.is_running:
                prog.start(True)

    def stop(self):
        pass



"""
Install a win32 service using the win32api
"""
import getpass
import logging
import os
from os.path import abspath
import sys

from pywintypes import error as Win32Error
import win32api, win32serviceutil, win32service

from chalmers import errors
from chalmers.scripts import service as service_script

log = logging.getLogger(__name__)

ERR_SERVICE_EXISTS = 1073

def InstallService(serviceName, displayName, startType=win32service.SERVICE_DEMAND_START,
                    serviceDeps=None,
                    errorControl=win32service.SERVICE_ERROR_NORMAL,
                    userName=None, password=None,
                    description=None):
    """
    This is a copy of a more advanced usecase.
    For chalmers serviceName and displayName are required and the
    defaults should be sufficient
    """
    serviceType = win32service.SERVICE_WIN32_OWN_PROCESS

    script = abspath(service_script.__file__)

    if script.endswith('.pyc') or script.endswith('.pyo'):
        script = script[:-1]

    commandLine = "%s %s" % (sys.executable, script)

    hscm = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_ALL_ACCESS)

    try:
        hs = win32service.CreateService(hscm,
                                serviceName,
                                displayName,
                                win32service.SERVICE_ALL_ACCESS,  # desired access
                    serviceType,  # service type
                    startType,
                    errorControl,  # error control type
                    commandLine,
                    None,
                    0,
                    serviceDeps,
                    userName,
                    password)

        if description is not None:
            try:
                win32service.ChangeServiceConfig2(hs,
                                                  win32service.SERVICE_CONFIG_DESCRIPTION,
                                                  description)
            except NotImplementedError:
                pass  # # ChangeServiceConfig2 and description do not exist on NT

        win32service.CloseServiceHandle(hs)
    finally:
        win32service.CloseServiceHandle(hscm)

def get_service_name(userName):
    "Return the service name for the given user"
    simpleUserName = userName.rsplit('\\', 1)[-1]
    svc_name = 'chalmers:manager:%s' % simpleUserName
    return svc_name

def is_running(userName=getpass.getuser()):
    "Test if the chalmers sevice is running for 'userName'"
    svc_name = get_service_name(userName)
    try:
        status = win32serviceutil.QueryServiceStatus(svc_name)
    except win32api.error:
        return False
    return status[1] == win32service.SERVICE_RUNNING

def is_installed(userName=getpass.getuser()):
    "Test if the chalmers sevice is installed for 'userName'"
    svc_name = get_service_name(userName)
    try:
        win32serviceutil.QueryServiceStatus(svc_name)
        return True
    except win32api.error:
        return False

def instart(userName, password):
    ''' Install and  Start (auto) a the chalmers service
    '''

    simpleUserName = userName.rsplit('\\', 1)[-1]
    svc_name = get_service_name(userName)
    display_name = 'Chalmers service manager for user %s' % simpleUserName

    win32api.SetConsoleCtrlHandler(lambda x: True, True)
    try:
        InstallService(
            svc_name,
            display_name,
            startType=win32service.SERVICE_AUTO_START,
            userName=userName,
            password=password,
        )

    except Win32Error as err:
        if err.args[0] != ERR_SERVICE_EXISTS:
            raise
        log.warn('Chalmers service %s is already installed' % svc_name)
    else:
        log.info('Installed chalmers service %s to windows services' % svc_name)

    try:
        win32serviceutil.StartService(svc_name)
    except Win32Error as err:
        log.error('StartService: %s' % err.args[2])
        if err.args[0] == 1069:
            log.error("This is usually a password error, please uninstall the service and retry")
            log.error(" *OR*")
            log.error("Run the command:\n\tsc config %s password= \"pass\"\nsc start %s" %
                      (svc_name, svc_name))
            log.error(" *OR*")
            log.error("Open the 'Services' application and edit it there")
        else:
            AllUsersProfile = os.environ.get('AllUsersProfile', 'C:\\ProgramData')
            logfile = os.path.join(AllUsersProfile, '%s-chalmers-service-log.txt' %
                                   userName.rsplit('\\', 1)[-1])
            log.error("Could not start the chalmers windows service for user %s" %
                      userName)
            log.error('Check the logfile "%s" for more details' % logfile)
    else:
        log.info('Start OK')





import unittest

from chalmers.service.cron_service import CronService, check
from contextlib import contextmanager
from chalmers import errors
import subprocess as sp
crontab_lines = None

def mock_check_output(*args, **kwargs):
    if crontab_lines is None:
        raise sp.CalledProcessError(1, 'crontab', 'crontab: no crontab for test')
    return crontab_lines

def mock_communicate(input):
    global crontab_lines
    crontab_lines = input
    return ''

class Test(unittest.TestCase):

    @contextmanager
    def mock_crontab(self, exists=True):
        from mock import patch
        patch1 = patch('subprocess.check_output')
        patch2 = patch('subprocess.Popen')
        with patch1 as check_output, patch2 as Popen:
            if  exists:
                check_output.side_effect = mock_check_output
                Popen().communicate.side_effect = mock_communicate
            else:
                check_output.side_effect = IOError(2, "err")
            yield

    def test_install(self):

        with self.mock_crontab(exists=True):

            service = CronService(False)
            self.assertFalse(service.status())
            self.assertTrue(service.install())
            self.assertTrue(service.status())
            self.assertTrue(service.uninstall())
            self.assertFalse(service.status())

    def test_check(self):
        global crontab_lines

        with self.mock_crontab(exists=False):
            self.assertFalse(check())

        with self.mock_crontab(exists=True):
            self.assertTrue(check())

            crontab_lines = 'some content'

            self.assertTrue(check())






if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_status']
    unittest.main()



import unittest
from chalmers.utils.persistent_dict import PersistentDict
import os

class Test(unittest.TestCase):

    def test_init(self):
        filename = 'test_filename.yaml'
        self.addCleanup(lambda : os.path.isfile(filename) and os.unlink(filename))
        d = PersistentDict(filename)
        d.update(x=1)

        self.assertEqual(d, {'x':1})

        d2 = PersistentDict(filename)

        self.assertEqual(d2, {'x':1})

    def test_update(self):
        filename = 'test_filename.yaml'
        self.addCleanup(lambda : os.path.isfile(filename) and os.unlink(filename))
        d = PersistentDict(filename)
        d.update(z=1)
        self.assertEqual(d, {'z':1})
        d.update(z=2, y=2)
        self.assertEqual(d, {'z':2, 'y':2})

        d2 = PersistentDict(filename)

        self.assertEqual(d2, d)

    def test_setitem(self):
        filename = 'test_filename.yaml'
        self.addCleanup(lambda : os.path.isfile(filename) and os.unlink(filename))
        d = PersistentDict(filename)
        d['l'] = 1
        self.assertEqual(d, {'l':1})

        d2 = PersistentDict(filename)
        self.assertEqual(d2, d)

if __name__ == '__main__':
    unittest.main()




'''
Uitility for deamonizing linux processes
'''
import logging
import os
import sys


log = logging.getLogger(__name__)

def daemonize(target, stream=None):
    """
    do the UNIX double-fork magic, see Stevens' "Advanced 
    Programming in the UNIX Environment" for details (ISBN 0201563177)
    http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
    """
    pid = os.fork()
    if pid > 0:
        # exit first parent
        return

    # decouple from parent environment
    os.chdir("/")
#     os.setsid()
#     os.umask(0)

    # do second fork
    pid = os.fork()
    if pid > 0:
        # exit from second parent
        sys.exit(0)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()

    si = open('/dev/null', 'r')
    so = open('/dev/null', 'a+')
    se = open('/dev/null', 'a+')

    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    if stream:
        sys.stdin = stream
        sys.stderr = stream

    # Run function as daemon
    target()
    sys.exit(0)



import unittest

from chalmers.service.darwin_service import DarwinService
from contextlib import contextmanager
import subprocess as sp
import xml.etree.ElementTree as ET

loaded_launchd = {}

def plist_value(value):
    if value.tag == 'string':
        return value.text
    elif value.tag == 'true':
        return True
    elif value.tag == 'false':
        return False
    elif value.tag == 'array':
        return [plist_value(v) for v in value.getchildren()]
    else:
        raise Exception("unexpected tag %s" % value)

def read_plist(path):
    etree = ET.parse(path)
    elems = etree.getroot().find('dict').getchildren()
    result = {}
    for i in range(0, len(elems), 2):
        key = elems[i].text
        result[key] = plist_value(elems[i + 1])
    return result

def mock_check_output(command, *args, **kwargs):
    if command[:2] == ['launchctl', 'list']:
        if command[2] not in loaded_launchd:
            raise sp.CalledProcessError(1, 'launchctl', 'not found')
        return loaded_launchd[command[2]]

    elif command[:2] == ['launchctl', 'load']:

        plist = read_plist(command[2])
        loaded_launchd[plist['Label']] = plist
        return repr(plist)
    elif command[:2] == ['launchctl', 'remove']:
        if command[2] not in loaded_launchd:
            raise sp.CalledProcessError(1, "launchctl", "no loaded")
        del loaded_launchd[command[2]]
        return 'ok'
    else:
        raise Exception('unknown command %r' % command)

class Test(unittest.TestCase):

    @contextmanager
    def mock_launchctl(self, exists=True):
        from mock import patch
        patch1 = patch('subprocess.check_output')
        with patch1 as check_output:
            if  exists:
                check_output.side_effect = mock_check_output
            else:
                check_output.side_effect = IOError(2, "err")
            yield

    def test_install(self):

        with self.mock_launchctl(exists=True):

            service = DarwinService(False)
            self.assertFalse(service.status())
            self.assertTrue(service.install())
            self.assertTrue(service.status())
            self.assertTrue(service.uninstall())
            self.assertFalse(service.status())
            self.assertFalse(service.uninstall())



if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_status']
    unittest.main()



"""
Install "SysV-style" init services definition with chkconfig command

http://en.wikipedia.org/wiki/Init

Note: chkconfig only soft links the /etc/inid.d/ script to the proper /etc/rc files.
We could check if this could be done manually. chkconfig behaves strange on some systems.
"""

from __future__ import unicode_literals, print_function

import logging
from os import path
import os
from subprocess import check_call, CalledProcessError, PIPE, check_output as _check_output
import sys

from chalmers.service.cron_service import CronService
import platform


log = logging.getLogger('chalmers.service')

INIT_D_DIR = '/etc/init.d'
python_exe = sys.executable
chalmers_script = sys.argv[0]

def read_data(filename):
    filename = path.join(path.dirname(__file__), 'data', filename)
    with open(filename) as fd:
        return fd.read()

def check():
    try:
        check_call(['chkconfig'], stdout=PIPE)
        return True
    except OSError as err:
        if err.errno == 2:
            return False
        raise

class SysVService(object):
    # Returns CronService if user is not root
    __new__ = CronService.use_if_not_root

    def __init__(self, target_user):
        self.target_user = target_user

        log.info('Platform: %s' % (platform.linux_distribution()[0] or 'Unknown'))
        log.info('Using Linux sysv chkconfig')
        if target_user:
            log.info('Chalmers service for target user: {}'.format(target_user))
        else:
            log.info('Chalmers service for root user')

    @property
    def script_name(self):
        if self.target_user:
            return 'chalmers.%s' % self.target_user
        else:  # Run as root
            return 'chalmers'


    @property
    def script_path(self):
        return path.join(INIT_D_DIR, self.script_name)

    @property
    def launch(self):
        if self.target_user:
            return '/bin/su - %s' % self.target_user
        else:  # Run as root
            return '/bin/sh'

    def check_output(self, command):
        log.info('Running command: %s' % ' '.join(command))
        output = _check_output(command)
        log.info(output)
        return output


    @property
    def template(self):
        return read_data('sysv-init.sh')


    def install(self):

        data = self.template.format(python_exe=python_exe,
                           chalmers=chalmers_script,
                           launch=self.launch,
                           script_name=self.script_name)


        with open(self.script_path, 'w') as fd:
            fd.write(data)

        log.info('Write file: %s' % self.script_path)
        os.chmod(self.script_path, 0o754)
        log.info('Running command chmod 754 %s' % self.script_path)

        # Redhat, CentOS
        self.check_output(['chkconfig', self.script_name, 'on'])


    def uninstall(self):
        self.check_output(['chkconfig', '--del', self.script_name])


        if not path.exists(self.script_path):
            log.warn("File '%s' does not exist " % self.script_path)
        else:
            os.unlink(self.script_path)

        log.info("Chalmers service has been removed")

    def status(self):

        try:
            self.check_output(['chkconfig', '--list', self.script_name])
        except CalledProcessError as err:
            if err.returncode == 1:
                log.info("Chalmers will not start on boot")
                return False
            raise


        if not path.exists(self.script_path):
            log.warn("Service file '%s' does not exist " % self.script_path)
        log.info("Chalmers is setup to start on boot")
        return True






'''
[Advanced] [In-development]

Export a program list to a single yaml file.

The export may contain machine specific paths.
and may need to be edited for portability
'''

from __future__ import unicode_literals, print_function

from argparse import FileType
import logging
import sys

import yaml

from chalmers.utils.cli import add_selection_group, select_programs


log = logging.getLogger('chalmers.export')



def main(args):

    export_data = []

    programs = select_programs(args, filter_paused=False)

    for prog in programs:
        export_data.append({'program': dict(prog.raw_data)})

    yaml.safe_dump(export_data, args.output, default_flow_style=False)

def add_parser(subparsers):
    parser = subparsers.add_parser('export',
                                      help='[IN DEVELOPMENT] Export current configuration to be installed with the "import" command',
                                      description=__doc__)

    add_selection_group(parser)

    parser.add_argument('-o', '--output', type=FileType('w'), default=sys.stdout)
    parser.set_defaults(main=main)



'''
Edit a program definition

Open the program definition file with the 
EDITOR environment variable

for example:
    
    chalmers edit server1
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import os
import pipes
import subprocess

from chalmers import errors
from chalmers.program import Program


log = logging.getLogger('chalmers.edit')


def main(args):

    EDITOR = os.environ.get('EDITOR', None)
    if EDITOR is None:
        raise errors.ChalmersError("Environment variable 'EDITOR' needs to be set")

    prog = Program(args.name, force=True)

    cmd = '%s %s' % (EDITOR, pipes.quote(prog.raw_data.filename))

    print(cmd)
    if subprocess.call(cmd, shell=True):
        raise errors.ChalmersError('Command "%s" exited with non zero status' % cmd)

    if prog.is_running:
        log.info("Changes to program %s will take effect on the next restart" % prog.name)

def add_parser(subparsers):
    parser = subparsers.add_parser('edit',
                                      help='Edit a service definition',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('name')
    parser.set_defaults(main=main)



from __future__ import print_function, unicode_literals
import io
import logging
import os
import shutil
import unittest

import mock
import yaml

from chalmers import config, errors
from chalmers.scripts import chalmers_main


class ChalmersCli(object):
    def __init__(self):
        self.script = chalmers_main.__file__
        if self.script.endswith('.pyc') or self.script.endswith('.pyo'):
            self.script = self.script[:-1]
        self.env = os.environ.copy()
        self.root = 'test_config'
        self.env['CHALMERS_ROOT'] = self.root
        config.set_relative_dirs(self.root)

        logging.getLogger('cli-logger').addHandler(logging.NullHandler())


    def __getattr__(self, subcommand):

        def run_subcommand(*args):
            cmd = ['-q', '--no-color', subcommand]
            cmd.extend(args)

            out = io.StringIO()

            log = logging.getLogger('chalmers')
            log.setLevel(logging.ERROR)
            del log.handlers[:]

            with mock.patch('sys.stdout', out), mock.patch('chalmers.scripts.chalmers_main.setup_logging'):
                chalmers_main.main(cmd, False)

            return out.getvalue()

        return run_subcommand


class Test(unittest.TestCase):

    def setUp(self):
        self.cli = ChalmersCli()
        if os.path.isdir(self.cli.root):
            shutil.rmtree(self.cli.root)
        unittest.TestCase.setUp(self)

    def tearDown(self):
        if os.path.isdir(self.cli.root):
            shutil.rmtree(self.cli.root)
        unittest.TestCase.tearDown(self)

    def test_add_show(self):
        self.cli.add('echo', 'hi', '--off', '--off')
        self.assertEqual(self.cli.show('echo', '--cmd').strip(), 'echo hi')
        self.assertEqual(self.cli.show('echo', '--cmd').strip(), 'echo hi')


    def test_cant_add_twice(self):
        self.cli.add('echo', 'hi', '--off')
        with self.assertRaises(errors.ChalmersError):
            self.cli.add('echo', 'hi', '--off')

    def test_add_remove(self):
        self.cli.add('echo', 'hi', '--off')
        self.cli.remove('echo')

    def test_cant_remove_non_existent(self):

        with self.assertRaises(SystemExit):
            self.cli.remove('echo')

    def test_cant_start_non_existent(self):

        with self.assertRaises(SystemExit):
            self.cli.start('echo')


    def test_list_no_programs(self):

        out = self.cli.list()
        self.assertEqual(out.strip(), "No programs added")

    def test_list(self):

        out = self.cli.add('echo', 'hi', '--off')
        out = self.cli.list()
        self.assertEqual(out.split(), ['echo', 'OFF'])

    def test_pause(self):

        out = self.cli.add('echo', 'hi', '--off')

        out = self.cli.list()
        self.assertEqual(out.split(), ['echo', 'OFF'])

        out = self.cli.on('echo')

        out = self.cli.list()
        self.assertEqual(out.split(), ['echo', 'STOPPED'])

        out = self.cli.off('echo')

        out = self.cli.list()
        self.assertEqual(out.split(), ['echo', 'OFF'])

    def test_show(self):

        self.cli.add('echo', 'hi', '--off')
        out = self.cli.show('echo')

    def test_set(self):

        self.cli.add('echo', 'hi', '--off')
        out = self.cli.set('echo', 'x=1')
        self.assertEqual(out.strip(), "Set 'x' to 1 for program echo\ndone")
        out = self.cli.show('echo', '--def')
        with open(out.strip()) as fd:
            data = yaml.load(fd)

        self.assertEqual(data['x'], 1)

    def test_start(self):

        self.cli.add('echo', 'This is the output', '--off')
        out = self.cli.start('echo', '-w')

    def test_log(self):

        self.cli.add('echo', 'This is the output', '--off')
        out = self.cli.start('echo', '-w')
        out = self.cli.log('echo')
        self.assertIn('This is the output', out)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Install a launchd rule to run at boot

http://en.wikipedia.org/wiki/Launchd
 
"""
from __future__ import unicode_literals, print_function

import logging
from os import path
import os
import subprocess as sp
import sys

from chalmers import errors

import tempfile

python_exe = sys.executable
chalmers_script = sys.argv[0]


def read_data(filename):
    filename = path.join(path.dirname(__file__), 'data', filename)
    with open(filename) as fd:
        return fd.read()

launchd_label = "org.continuum.chalmers"

log = logging.getLogger('chalmers.service')

class DarwinService(object):
    def __init__(self, target_user):
        self.target_user = target_user

        log.info('Platform: Darwin')
        log.info('Using Darwin launchd service')

        if target_user:
            log.info('Launching chalmers as target user %s' % target_user)
        elif target_user is False:
            log.info('Launching chalmers as current user (does not require root)')
        else:
            log.info('Launching chalmers as root user')

    @property
    def label(self):
        if self.target_user:
            return '%s.%s' % (launchd_label, self.target_user)
        else:
            return launchd_label

    @property
    def template(self):
        return read_data('launchd.plist')

    def check_output(self, command):
        if self.target_user:
            if os.getuid() != 0:
                raise errors.ChalmersError("Can not perform system install without root")

        log.info("Running command: %s" % ' '.join(command))
        try:
            output = sp.check_output(command, stderr=sp.STDOUT)
        except OSError as err:
            raise errors.ChalmersError("Could not access program 'launchctl' required for osx service install")
        except sp.CalledProcessError as err:
            if err.returncode == 1:
                if 'Socket is not connected' in err.output:
                    log.error(err.output)
                    raise errors.ChalmersError("The user '%s' must be logged in via the osx gui to perform this operation" % self.target_user)
            raise

        return output

    def get_launchd(self):
        try:
            command = ['launchctl', 'list', self.label]
            return self.check_output(command)
        except sp.CalledProcessError as err:
            if err.returncode == 1:
                return None
            raise

    def add_launchd(self):
        if self.target_user:
            username = '<key>UserName</key> <string>%s</string>' % self.target_user
        else:
            username = ''
        plist = self.template.format(python_exe=python_exe,
                                     chalmers=chalmers_script,
                                     label=self.label,
                                     username=username)

        with tempfile.NamedTemporaryFile('w', suffix='.plist', prefix='chalmers') as fd:

            fd.write(plist)
            fd.flush()
            try:
                command = ['launchctl', 'load', fd.name]
                self.check_output(command).strip()
            except sp.CalledProcessError as err:
                if err.returncode == 1:
                    raise errors.ChalmersError("Chalmers service is already installed")
                raise


    def install(self):
        """Create a launchd plist and load as a global daemon"""

        log.info("Adding chalmers launchd plist")
        self.add_launchd()
        log.info("All chalmers programs will now run on boot")
        return True

    def uninstall(self):
        """Uninstall launchd plist for chalmers"""

        log.info("Removing chalmers plist from launchd")
        try:
            command = ['launchctl', 'remove', self.label]
            self.check_output(command).strip()
        except sp.CalledProcessError as err:
            if err.returncode == 1:
                log.error("Chalmers service is not installed")
                return False
            raise

        log.info("Chalmers service has been removed")
        return True

    def status(self):
        """Check if chalmers will be started at reboot"""
        launchd_lines = self.get_launchd()
        if launchd_lines:
            log.info("Chalmers is setup to start on boot")
            return True
        else:
            log.info("Chalmers will not start on boot")
            return False




import logging
import psutil

log = logging.getLogger(__name__)

def kill_tree(pid):
    'Kill all processes and child processes'

    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        log.error("NoSuchProcess pid=%s" % pid)
        return

    children = parent.children(recursive=True)

    parent.kill()

    for child in children:
        if child.is_running():
            child.kill()



'''
Enable/Disable chalmers so that it will run at user login.

This command does not require root/admin access for all platforms.

This command will install chalmers server to start at
login for the current user (posix)::

    chalmers @login enable

If you want to enable chalmers as a system wide service try::
    
    chalmers @startup --help
'''

from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import os

from chalmers import errors
from chalmers.service import LocalService


log = logging.getLogger(__name__)

def main(args):

    service = LocalService(False)

    if args.action == 'status':
        service.status()
    elif args.action == 'enable':
        service.install()
    elif args.action == 'disable':
        service.uninstall()

    else:
        raise errors.ChalmersError("Invalid action %s" % args.action)


def add_parser(subparsers):
    parser = subparsers.add_parser('@login',
                                   help='Install chalmers as a service',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('action', choices=['enable', 'disable', 'status'], nargs='?', default='status')
    group = parser.add_argument_group('Service Type').add_mutually_exclusive_group()

    if os.name == 'posix':
        sytem_default = os.environ.get('SUDO_USER') if os.getuid() == 0 else None
    else:
        sytem_default = None

    group.add_argument('--user', dest='target_user', default=sytem_default, metavar='USERNAME',
                       help='Install Chalmers as a service to the system for a given user (requires admin). '
                            'If no user is given it will launch chalmers as root (default: %(default)s)')

    parser.set_defaults(main=main)




'''List the all the defined programs and their state

eg: 

    $ chalmers list
    server1                             RUNNING         pid 19278 , uptime 0:10:01

    
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
from datetime import timedelta
import logging
import time

from chalmers.program import Program
import sys
import os
import yaml


log = logging.getLogger(__name__)

def sort_key(prog):
    return prog.text_status, prog.name


def list_loaded(args):

    programs = sorted(Program.find_for_user(), key=sort_key)

    for prog in programs:

        if prog.is_running:
            start_time = prog.state.get('start_time')
            if start_time:
                td = str(timedelta(seconds=(time.time() - start_time))).rsplit('.', 1)[0]
            else:
                td = '?'
            ctx = (prog.data['name'][:24], prog.text_status,
                   prog.state.get('child_pid'),
                   td,
                   )

            print('%-25s %-15s pid %-6s, uptime %s' % ctx)
        else:
            stop_time = prog.state.get('stop_time')
            if stop_time:
                td = 'downtime ' + str(timedelta(seconds=(time.time() - stop_time))).rsplit('.', 1)[0]
            else:
                td = ''

            reason = prog.state.get('reason', '')
            ctx = (prog.data['name'][:35], prog.text_status,
                   reason

                  )

            print('%-25s %-15s %s' % ctx)

            if td:
                print(' ' * 41, td)

    if not programs:
        print('No programs added')


def list_stashed(args):

    locations = [os.path.join(sys.prefix, 'etc', 'chalmers'),
                 os.path.join('/', 'etc', 'chalmers')]
    for location in locations:
        if not os.path.isdir(location):
            continue
        print(location)
        for filename in os.listdir(location):
            filepath = os.path.join(location, filename)
            print(filepath)
            print(yaml.load(open(filepath)))


def main(args):
    if args.stashed:
        list_stashed(args)
    else:
        list_loaded(args)

def add_parser(subparsers):
    description = 'List registered programs'
    parser = subparsers.add_parser('list',
                                   help=description, description=description,
                                   epilog=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('--stashed', action='store_true')
    parser.set_defaults(main=main)





'''
Enable/Disable chalmers so that it will run at machine startup.

This command requires root/admin access for all platforms.

If you need to enable chalmers as a service without admin try::

    chalmers @login --help

This command will install chalmers server to start on
boot for the current user (posix)::

    sudo chalmers @starup enable

This command will install chalmers server to start on
boot for a target user (posix)::

    sudo chalmers service install --root

Root service install (windows)::
    
    runas /user:.\Administrator "chalmers @startup enable"
'''

from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import os

from chalmers import errors
from chalmers.service import SystemService


def main(args):


    if os.name == 'nt':
        from win32com.shell import shell
        if not shell.IsUserAnAdmin():
            raise errors.ChalmersError('You must be an administrator to run this command')
    else:
        if os.getuid() != 0:
            raise errors.ChalmersError('You must be root to run this command')

    service = SystemService(args.target_user)

    if args.action == 'status':
        service.status()
    elif args.action == 'enable':
        service.install()
    elif args.action == 'disable':
        service.uninstall()

    else:
        raise errors.ChalmersError("Invalid action %s" % args.action)


def add_parser(subparsers):
    parser = subparsers.add_parser('@startup',
                                   help='Install chalmers as a service',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('action', choices=['enable', 'disable', 'status'], nargs='?', default='status')
    group = parser.add_argument_group('Service Type').add_mutually_exclusive_group()

    if os.name == 'posix':
        sytem_default = os.environ.get('SUDO_USER') if os.getuid() == 0 else None
    else:
        sytem_default = None

    group.add_argument('--user', dest='target_user', default=sytem_default, metavar='USERNAME',
                       help='Install Chalmers as a service to the system for a given user (requires admin). '
                            'If no user is given it will launch chalmers as root (default: %(default)s)')

    parser.set_defaults(main=main)




import unittest
from chalmers.event_dispatcher import EventDispatcher
from chalmers import config, errors

from os.path import join
import tempfile
import os

class TestEventDispatcher(EventDispatcher):

    @property
    def name(self):
        return 'test'

    def dispatch_foo(self, value):
        self.foo = value

    def dispatch_error(self, value):
        raise Exception("Expected dispatch error")

class Test(unittest.TestCase):

    def setUp(self):

        self.root_config = join(tempfile.gettempdir(), 'chalmers_tests')
        config.set_relative_dirs(self.root_config)

        unittest.TestCase.setUp(self)


    def test_init(self):
        dispatcher = TestEventDispatcher()
        self.assertFalse(dispatcher.is_listening)

    def test_listen(self):

        d = TestEventDispatcher()

        if os.path.exists(d.addr):
            os.unlink(d.addr)

        self.assertFalse(d.is_listening)

        d.start_listener()

        self.assertTrue(d.is_listening)

        if os.name != 'nt':
            self.assertTrue(os.path.exists(d.addr))

        d.send('foo', 1)
        d.send('exitloop')

        d._listener_thread.join()

        self.assertFalse(d.is_listening)
        self.assertEqual(d.foo, 1)

    def test_exception(self):

        d = TestEventDispatcher()

        if os.path.exists(d.addr):
            os.unlink(d.addr)

        self.assertFalse(d.is_listening)

        d.start_listener()

        with self.assertRaises(errors.ChalmersError):
            d.send('error', 1)

        self.assertTrue(d.is_listening)

        d.send('exitloop')

        d._listener_thread.join()

        self.assertFalse(d.is_listening)




if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
TODO: this may also be obsolete

    This could also replace some of the functionality in `commands/*` and
    let them be more about user interaction.
"""
from contextlib import contextmanager
import itertools
import logging
from multiprocessing import Process, Manager
import random
import sys

from clyent.logs.colors import color
from clyent.logs.formatters import FormatterWrapper
from clyent.logs.handlers import ColorStreamHandler

from chalmers.event_dispatcher import EventDispatcher
from chalmers.program import Program
from clyent.logs import log_unhandled_exception


log = logging.getLogger(__name__)

class ProgramManager(EventDispatcher):
    """
    Manages chalmers programs with multiprocessing
    
    Listens to 'start' events to start new programs
    """

    NAME = 'chalmers_manager'

    COLOR_CODES = list(range(40, 48)) + [100, 102, 104, 105, 106]
    random.shuffle(COLOR_CODES)

    def __init__(self, exit_on_first_failure=False, use_color=None, setup_logging=True):
        EventDispatcher.__init__(self)
        self.manager = Manager()
        self.processes = []

        self.setup_logging = setup_logging
        if use_color is None:
            use_color = sys.stdout.isatty()

        self.use_color = use_color
        self.bg_colors = itertools.cycle(color.BACKGROUND_COLORS)
        self.exit_on_first_failure = exit_on_first_failure


    @property
    def name(self):
        return self.NAME

    def dispatch_start(self, name):
        log.info("Managing Program %s" % name)
        p = Process(target=start_program,
                    name='start_program:%s' % name,
                    args=(name,),
                    kwargs={'color_id': self.use_color and next(self.bg_colors),
                            'setup_logging':self.setup_logging})

        p.start()
        self.processes.append(p)

    def start_all(self):
        for prog in Program.find_for_user():
            if not prog.is_paused:
                self.dispatch_start(prog.name)
            else:
                log.info("Not starting program %s (it is paused)" % (prog.name))

    @contextmanager
    def cleanup(self, prog):
        try:
            yield
        except KeyboardInterrupt:
            log.error('KeyboardInterrupt')
        finally:
            if not prog.is_ok:
                log.info("Program manager letting program fail")


def start_program(name, color_id=None, setup_logging=True):

    logger = logging.getLogger('chalmers')
    logger.setLevel(logging.INFO)

    prefix = '[%s]' % name
    if color_id is not None:
        prefix = color(prefix, [color_id, color.WHITE])

    if setup_logging and not logger.handlers:
        shndlr = ColorStreamHandler(color_id is not None)
        shndlr.setLevel(logging.INFO)
        logger.setLevel(logging.INFO)
        logger.addHandler(shndlr)

    if setup_logging:
        for h in logger.handlers:
            FormatterWrapper.wrap(h, prefix=prefix)



    sys.excepthook = log_unhandled_exception(logger)



    prog = Program(name)

    prog.start_sync()




"""
Chalmers error classes
"""
from clyent.errors import ClyentError

class ChalmersError(ClyentError):
    """
    Generic error that will hide the traceback
    unless --show-traceback is given

    the 'message' must then always be immediately obvious
    to the end user
    """
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else None
        ClyentError.__init__(self, *args, **kwargs)

class ProgramNotFound(ChalmersError):
    'No program definition found on disk'
    pass

class StateError(ChalmersError):
    pass

class ConnectionError(ChalmersError):
    pass

class StopProcess(Exception):
    'called to signal a process should exit'



"""
TODO: determine if this is still required
"""
from __future__ import print_function, unicode_literals, absolute_import

import getpass
import sys, os
import time
import traceback


def main():

    AllUsersProfile = os.environ.get('AllUsersProfile', 'C:\\ProgramData')
    logfile = os.path.join(AllUsersProfile, '%s-chalmers-service-log.txt' % getpass.getuser())

    try:
        logfd = open(logfile, 'a', 1)
    except:
        # I guess we will have to leave it up to the windows event log
        traceback.print_exc()
    else:
        sys.stdout = sys.stderr = logfd

    print('---')
    print("Starting Chalmers Service", time.ctime())
    sys.stdout.flush()
    try:
        from chalmers.windows.chalmers_service import ChalmersService
        import servicemanager
        servicemanager.Initialize(None, None)
        servicemanager.PrepareToHostSingle(ChalmersService)
        servicemanager.StartServiceCtrlDispatcher()
    finally:
        print("Exiting Chalmers Service", time.ctime())
        sys.stdout.flush()


if __name__ == "__main__":
    main()



'''
Allow a program to run in the background 

Example:
    From a terminal 'window 1' run::
    
      $ chalmers start --wait server1
    
    program output ...
    program output ...
    
    From another Another terminal 'window 2' run::
      
      $ chalmers bg server1
    
    Now you will be able to close the terminal 'window 1' without shutting down your server

'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging

from chalmers import errors
from chalmers.program import Program


log = logging.getLogger('chalmers.edit')


def main(args):

    prog = Program(args.name)

    if not prog.is_running:
        raise errors.StateError("Program is not running")

    prog.send('bg')
    print("Process is now a background process")

def add_parser(subparsers):
    description = 'Allow a service to run in the background'
    parser = subparsers.add_parser('bg',
                                   help=description, description=description,
                                   epilog=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('name')
    parser.set_defaults(main=main)



'''
Set a definition variable for a program

To update a programs environment or start configuration
 
Example - to change the signal used by 'chalmers stop' to shut down a process:

    chalmers set server1 stopsignal=SIGINT

Example - to change an environment variable:

    chalmers set server1 env.PORT=5000
    

Common Config values:

  * startsecs: The time in seconds that the program is assumed to be starting up
    If the program exits before this time it is considered to be spinning
  * startretries: The number or times to launch a spinning program
  * stopwaitsecs: Wait this long
  * exitcodes: A list of exit codes that are accepted as a successful exit 
  * stopsignal: The signal to sent to terminate the program. May be an int or string
    eg: 'SIGTERM' or 15  
  * cwd: Directory run the command in
  

Log file config values:

 * log_dir: The base directory to output logs
 * redirect_stderr: Direct stderr to the same log file as stdout
 * stdout: filename to pipe the program's stdout
 * stderr: filename to pipe the program's stderr
 * daemon_log: filename to pipe the programs conrol log 
 * env.PYTHONUNBUFFERED: Set this value to 1 if you want are running a 
   python program and want realtime logging 
   See: https://docs.python.org/2/using/cmdline.html#envvar-PYTHONUNBUFFERED 

Posix Only Config values:

  * umask: Abbreviation of user mask: sets the file mode creation mask of the current process. 
   See http://en.wikipedia.org/wiki/Umask
  * user: User to run the program as. May be a username or UID. This option is only valid when 
    chalmers is run as the root user 
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging

from chalmers import errors
from chalmers.program import Program
from chalmers.utils import try_eval, set_nested_key


log = logging.getLogger('chalmers.set')

def split_try_eval(item):
    'split key=value pairs, then eval the value'
    if '=' not in item:
        raise TypeError("Chalmers can not set '%s' as a definition variable\n"
                        "The parameter must contain an '='" % (item))
    key, value = item.split('=', 1)
    return key, try_eval(value)

def main(args):

    proc = Program(args.name)

    if proc.is_running:
        log.warning("Program is running: Updates will not be reflected until a restart is done")
    with proc.raw_data.transaction():
        for key, value in args.items:
            if key == 'name':
                raise errors.ChalmersError("Can not set program name")

            set_nested_key(proc.raw_data, key, value)
            print("Set '%s' to %r for program %s" % (key, value, args.name))

    print("done")


def add_parser(subparsers):
    parser = subparsers.add_parser('set',
                                      help='Set a variable in the program definition',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('name', metavar='PROG',
                        help='The name of the program to update')
    parser.add_argument('items', nargs='+', metavar='KEY=VALUE', type=split_try_eval,
                        help='Update the program with these key-value pairs')
    parser.set_defaults(main=main)



"""
chalmes global config vars
"""
from __future__ import unicode_literals, print_function

from chalmers.utils import appdirs
import os


dirs = appdirs.AppDirs('chalmers', 'srossross')

def set_relative_dirs(root):
    'Set the application directory relative root'
    global dirs
    dirs = appdirs.RelativeAppDirs(os.path.abspath(root))

def main_logfile():
    return os.path.join(dirs.user_log_dir, 'chalmers.log')

# Set the root config and log dirs from an environment variable
if os.environ.get('CHALMERS_ROOT'):
    CHALMERS_ROOT = os.environ['CHALMERS_ROOT']
    set_relative_dirs(CHALMERS_ROOT)



'''
Show log output of program:

To show the stdout of a run::

    chalmers log server1

To show the chalmers program control log, run::

    chalmers log server1 -3

To show the global chalmers deamon log run::

    chalmers log


You can allso  follow a file (like `tail -f`)

    chalmers log -f
'''

from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import os
import sys
import time

from chalmers import config
from chalmers.program import Program


log = logging.getLogger('chalmers.log')

def main(args):
    if args.name:
        prog = Program(args.name)

        logfile = prog.data.get(args.logfile)
    else:
        logfile = config.main_logfile()
        prog = None


    if args.showfile:
        print(logfile)
        return
    if args.rm:
        if os.path.isfile(logfile):
            log.info("Removing log file '%s'" % logfile)
            os.unlink(logfile)
        else:
            log.warn("File '%s' does not exist" % logfile)
        return

    if logfile is None:
        log.warn("The logfile for key '%s' does not exist" % args.logfile)
        return

    with open(logfile) as fd:
        if args.n:
            lines = fd.readlines()
            if args.n > 0:
                lines = lines[:args.n]
            else:
                lines = lines[args.n:]

            sys.stdout.write(''.join(lines))

        else:
            sys.stdout.write(fd.read())

        while args.append and (not prog or prog and prog.is_running):
            pos = fd.tell()
            if pos < os.fstat(fd.fileno()).st_size:
                fd.seek(pos)
                sys.stdout.write(fd.read())
            else:
                time.sleep(1)


def add_parser(subparsers):
    parser = subparsers.add_parser('log',
                                      help='Show program output',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('name', nargs='?')
    parser.add_argument('-f', action='store_true', dest='append',
                       help='The -f option causes log to not stop when end of file is reached, '
                            'but rather to wait for additional data to be appended')

    parser.add_argument('--tail', type=lambda arg:-int(arg), dest='n', metavar='N',
                        default=-10,
                       help='Tail lines at end of file (default: %(default)s)',)
    parser.add_argument('--head', type=int, dest='n', metavar='N',
                       help='Print only first lines of the file')

    parser.add_argument('--showfile', action='store_true',
                       help='Print the name of the logfile')
    parser.add_argument('--rm', '--remove', action='store_true',
                       help='Remove the logfile')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-1', '--stdout', action='store_const', const='stdout', dest='logfile',
                       default='stdout',
                       help='Show program stdout')
    group.add_argument('-2', '--stderr', action='store_const', const='stderr', dest='logfile',
                       help='Show program stderr')
    group.add_argument('-3', '--daemon', action='store_const', const='daemon_log', dest='logfile',
                       help='Show program manager log')

    parser.set_defaults(main=main)



'''
'''
from __future__ import print_function, unicode_literals
from yaml import safe_load
import logging
from pprint import pformat


log = logging.getLogger(__name__)

def try_eval(value):
    "Try and evaluate a literal value, if it fails, return a string"
    try:
        return safe_load(value)
    except:
        return value

def set_nested_key(dct, key, value):
    """
    Set a nested key in a dictionary
    """

    keys = key.split('.')
    next_key = keys.pop(0)

    while keys:
        dct = dct.setdefault(next_key, {})
        next_key = keys.pop(0)

    dct[next_key] = value
    pass

def pformat2(value):
    pvalue = pformat(value, width=66).split('\n')
    sep = '\n' + ' ' * 14
    return sep.join(pvalue)


def print_opts(category, data, opts, file=None):
    'Print all options from dict "data" that are in "category"'
    if not any(opt in data for opt in opts):
        return
    print('\n%s' % category, file=file)
    print('-' * len(category), file=file)
    for opt in opts:
        if opt in data:
            value = data.get(opt)
            print("%12s: %s" % (opt, pformat2(value)), file=file)



"""

"""
from __future__ import print_function, absolute_import, unicode_literals

from chalmers.utils.color_picker import ColorPicker
from chalmers.program import Program
import logging
from multiprocessing import Process, Queue
from os import path
from threading import Thread
import time


log = logging.getLogger(__name__)

SLEEP_START = 0.1
SLEEP_INC = 0.05
SLEEP_MAX = 1.0

def start_program(name):
    program = Program(name)
    # Remove base logging handlers, this is read from the log files that are
    # set up after the start method
    logger = logging.getLogger('chalmers')
    logger.handlers = []
    logger.handlers.append(logging.NullHandler())
    try:
        program.start(daemon=False)
    except KeyboardInterrupt:
        log.error('Program %s is shutting down' % program.name)

class MultiPlexIOPool(object):
    """
    This class runs programs in a sub-process optionally it watches their 
    stdout file outputs and multiplexes the output to the current processes stdout
    """

    def __init__(self, stream=False, use_color=False):
        self.stream = stream
        self.use_color = use_color

        self.processes = []
        self.programs = []

        self.queue = Queue()
        self.finished = False
        self.watched = []
        self.creating = []




    def printer_loop(self):

        colors = ColorPicker()
        sleep_time = SLEEP_START

        try:
            while not self.finished:
                seen_data = False
                for name, fd in list(self.watched):
                    data = fd.readline()
                    while data:
                        seen_data = True
                        if self.use_color:
                            with colors[name]:
                                print('[%s]' % name, end='')
                            print(" ", end='')
                        else:
                            print("[%s]" % name, end='')

                        print(data, end='')
                        data = fd.readline()

                if not seen_data:
                    time.sleep(sleep_time)
                    sleep_time = SLEEP_MAX if sleep_time >= SLEEP_MAX else sleep_time + SLEEP_INC
                else:
                    sleep_time = SLEEP_START

                self.manage_logs()

        finally:
            for _, fd in self.watched:
                try:
                    fd.close()
                except IOError: pass

    def manage_logs(self):
        if not self.creating:
            return
        name, filename = self.creating[-1]
        if path.isfile(filename):
            fd = open(filename)
            self.watched.append([name, fd])
            self.creating.pop()


    def add_log_streams(self, name, filename):

        if filename:
            if path.isfile(filename):
                fd = open(filename)
                fd.seek(0, 2)
                self.watched.append([name, fd])
            else:
                self.creating.append([name, filename])


    def append(self, program):

        proc = Process(target=start_program, args=(program.name,))
        self.programs.append(program)
        self.processes.append(proc)

        if self.stream:
            self.add_log_streams(program.name, program.data.get('daemon_log'))
            self.add_log_streams(program.name, program.data.get('stdout'))

        proc.start()


    def join(self):

        if self.stream:
            self.printer = Thread(target=self.printer_loop)
            self.printer.start()

        try:
            for proc in self.processes:
                proc.join()

        finally:
            self.finished = True


        log.info("All programs exited")




from setuptools import setup, find_packages
import os
import versioneer

install_requires = ['psutil>=3', 'clyent', 'pyyaml']

if os.name == 'nt':
    install_requires.append('pywin32')


setup(
    name='chalmers',

    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),

    author='Continuum Analytics',
    author_email='srossross@gmail.com',
    url='http://github.com/binstar/chalmers',
    description='Process Control System',
    packages=find_packages(),
    install_requires=install_requires,
    package_data={
       'chalmers.service': ['data/*'],
    },

    entry_points={
          'console_scripts': [
              'chalmers = chalmers.scripts.chalmers_main:main',
              ]
                 },
)




from __future__ import print_function, absolute_import, unicode_literals
import yaml
import os
from contextlib import contextmanager
from chalmers import errors

class PersistentDict(dict):
    """
    A class that persists a dict to a file

    This class behaves like a dict and adds new functionality to store the dict
    to a file when writing.
    """
    def __init__(self, filename, load=True):
        self._filename = os.path.abspath(filename)
        if load: self._load()
        self._transact = False

    @property
    def filename(self):
        'The filepath to write'
        return self._filename

    def exists(self):
        'test if the filename exists'
        return os.path.exists(self._filename)

    def reload(self):
        'perform a load/store'
        self._load()
        self._store()

    def delete(self):
        'remove the file from disk'
        if os.path.isfile(self._filename):
            os.unlink(self._filename)

    def _load(self):
        'load dict data into the current object'
        if os.path.isfile(self._filename):
            with open(self._filename) as fd:
                dict.clear(self)
                data = yaml.safe_load(fd)
                if data:
                    dict.update(self, data)

    def _store(self):
        'store dict data into the current object'
        dir = os.path.dirname(self._filename)
        if dir and not os.path.isdir(dir):
            os.makedirs(dir)

        with open(self._filename, 'w') as fd:
            yaml.safe_dump(dict(self), fd)


    def _mk_lockfile(self):
        'TODO: not implemented'
        lockfile = self._filename + '.lock'
        try:
            if self.exists():
                os.link(self._filename, lockfile)
            else:
                os.mkdir(lockfile)
        except OSError as err:
            if err.errno == 17:
                msg = ("The file '%s' is locked by another process.\n"
                       "If this process is not running, "
                       "you can manually remove the lockfile\n\trm '%s'" % (self._filename, lockfile))
                raise errors.ChalmersError(msg)
            raise

    def _rm_lockfile(self):
        'TODO: not implemented'
        lockfile = self._filename + '.lock'
        if os.path.isdir(lockfile):
            os.rmdir(lockfile)
        elif os.path.isfile(lockfile):
            os.unlink(lockfile)

    @contextmanager
    def file_lock(self):
        'TODO: not implemented'
        transact = self._transact
        self._transact = True

        if not transact:
            self._mk_lockfile()

        try:
            yield transact
        finally:
            if not transact:
                self._rm_lockfile()

        self._transact = transact

    @contextmanager
    def transaction(self):
        """
        Signify that an atomic transaction will occurr
        when the transaction context manager exits then
        all the data is stored to a file

        This is save to use in a nested way. the top level manager will be the only
        on that stores the data
        """
        # TODO: lock the file while the transation is going on
        # with self.file_lock() as transact:

        # Tell dict that a transation is going on
        transact = self._transact
        self._transact = True

        # don't load if already in a transation state
        if not transact:
            self._load()

        yield

        # don't store if already in a transation state
        # this means that the transaction context can be nested
        if not transact:
            self._store()

        # Restore previous state
        self._transact = transact


    #===============================================================================
    # Overload std dict setter methods
    #===============================================================================

    def __setitem__(self, *args, **kwargs):
        with self.transaction():
            return dict.__setitem__(self, *args, **kwargs)

    def update(self, *args, **kwargs):
        with self.transaction():
            return dict.update(self, *args, **kwargs)

    def setdefault(self, *args, **kwargs):
        with self.transaction():
            return dict.setdefault(self, *args, **kwargs)

    def pop(self, *args, **kwargs):
        with self.transaction():
            return dict.pop(self, *args, **kwargs)

    def popitem(self, *args, **kwargs):
        with self.transaction():
            return dict.popitem(self, *args, **kwargs)

    def clear(self):
        with self.transaction():
            return dict.clar(self)



'''
Start a program.  
The program must be defined with the chalmers 'run' command first

example:

    chalmers start server1
'''
from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import sys
import time

from clyent import print_colors

from chalmers import errors
from chalmers.utils import cli
from chalmers.utils.mutiplex_io_pool import MultiPlexIOPool
import os


log = logging.getLogger('chalmers.start')


def main(args):

    programs = cli.select_programs(args, filter_paused=True)

    programs = cli.filter_programs(programs, lambda p: p.is_running,
                               'Starting', 'started', True)

    if args.daemon:
        for prog in programs:
            if not prog.is_running:
                prog.start(args.daemon)

        for prog in programs:
            print("Checking status of program %-25s ... " % (prog.name[:25]), end='')
            sys.stdout.flush()
            err = prog.wait_for_start()
            if err:
                print_colors('[{=ERROR!c:red} ]')
            else:

                print_colors('[  {=OK!c:green}  ]')

    else:
        pool = MultiPlexIOPool(stream=args.stream, use_color=args.color and os.name == 'posix')

        for prog in programs:
            pool.append(prog)

        pool.join()


def restart_main(args):

    programs = cli.select_programs(args, filter_paused=True)

    if not (args.all or args.names):
        raise errors.ChalmersError("Must specify at least one program to restart")

    if len(programs):
        print("Restarting programs %s" % ', '.join([p.name for p in programs]))
        print("")
    else:
        log.warn("No programs to restart")
        return

    for prog in programs:

        sys.stdout.flush()

        if prog.is_running:
            print("Stop program %-25s ... " % (prog.name[:25]), end='')
            try:
                prog.stop()
            except errors.StateError:
                print_colors('[  {=ERROR!c:red}  ]')

            print_colors('[  {=OK!c:green}  ]')

    time.sleep(.5)

    for prog in programs:
        prog.start()

    for prog in programs:
        print("Checking status of program %-25s ... " % (prog.name[:25]), end='')
        sys.stdout.flush()
        err = prog.wait_for_start()
        if err:
            print_colors('[{=ERROR!c:red} ]')
        else:

            print_colors('[  {=OK!c:green}  ]')



def add_parser(subparsers):
    parser = subparsers.add_parser('start',
                                      help='Start a program',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    cli.add_selection_group(parser)

    parser.add_argument('-w', '--wait', '--no-daemon', action='store_false', dest='daemon',
                        help='Wait for program to exit')
    parser.add_argument('-d', '--daemon', action='store_true', dest='daemon', default=True,
                        help='Run program as daemon')
    parser.add_argument('--stream', '--io', action='store_true', default=sys.stdout.isatty(),
                        help='Multiplex stdout of programs to stdout')
    parser.add_argument('--no-stream', action='store_false', dest='stream',
                        help='Don\'t pip stdout of programs to stdout')

    parser.set_defaults(main=main)

    parser = subparsers.add_parser('restart',
                                      help='Restart a program',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter)

    cli.add_selection_group(parser)

    parser.set_defaults(main=restart_main)



import logging
import os
from threading import Thread
import time


log = logging.getLogger(__name__)

class FileEcho(Thread):
    """
    Perhaps obsolete class that echos the stdout logfile
    for a program
    """

    def __init__(self, filename, stream):
        self.filename = filename
        self.stream = stream
        self._running = True
        Thread.__init__(self, name='file_echo')
        self.daemon = True

    def stop(self):
        self._running = False

    def run(self):

        with open(self.filename, 'a+', 1) as fd:
            fd.seek(0, os.SEEK_END)
            pos = fd.tell()

            while self._running:
                if pos < os.fstat(fd.fileno()).st_size:
                    fd.seek(pos)
                    data = fd.read()
                    if isinstance(data, bytes):
                        data = data.decode('utf-8')
                    pos = fd.tell()
                    try:
                        self.stream.write(data)
                    except TypeError as err:
                        log.exception(err)
                else:
                    time.sleep(1)




"""
Contains client functions to help select and filter programs
from argparse
"""
import logging

from chalmers.program import Program
import sys


log = logging.getLogger(__name__)

def add_selection_group(parser):
    """
    Add program selection arguments
    """
    group = parser.add_argument_group("Select Programs")
    group.add_argument('names', nargs='*', metavar='PROG',
                        help='Names of the programs to select')
    group.add_argument('-a', '--all', action='store_true',
                        help='select all programs')


def select_programs(args, filter_paused=True, force=False):
    """
    Return a list of selected programs from command line arguments
    """

    if not (args.all ^ bool(args.names)):
        if args.all:
            log.error("You may not specify a program name when you use the -a/--all option (See -h/--help for more details)")
        else:
            log.error("You must select at least one program from the command line (See -h/--help for more details)")
        raise  SystemExit(1)

    if args.all:
        programs = list(Program.find_for_user(force=force))
        if filter_paused:
            programs = [prog for prog in programs if not prog.is_paused]

    else:
        programs = [Program(name, force=force) for name in args.names]

    erorrs = 0
    for program in programs:
        if not program.exists():
            erorrs += 1
            log.error("Program '%s' does not exist" % program.name)
    if erorrs:
        raise SystemExit(1)

    return list(programs)

def filter_programs(programs, filt, title, name, have_filter_paused=True):
    already_in_state = [p.name for p in programs if filt(p)]

    if already_in_state:
        log.warn("The programs '%s' are already %s" % ("', '".join(already_in_state), name))

    programs = [p for p in programs if not filt(p)]

    if len(programs):
        log.info("%s programs %s" % (title, ', '.join([p.name for p in programs])))
        log.info("")
    else:
        msg = "All "
        if have_filter_paused:
            msg += "paused "
        msg += "programs already %s" % (name)
        log.warn(msg)
        raise SystemExit(0)

    return programs


def bool_input(prompt, default=True):
    """
    Wait for bool input
    """
    default_str = '[Y|n]' if default else '[y|N]'
    while 1:
        inpt = input('%s %s: ' % (prompt, default_str))
        if inpt.lower() in ['y', 'yes'] and not default:
            return True
        elif inpt.lower() in ['', 'n', 'no'] and not default:
            return False
        elif inpt.lower() in ['', 'y', 'yes']:
            return True
        elif inpt.lower() in ['n', 'no']:
            return False
        else:
            sys.stderr.write('please enter yes or no\n')



'''
Stop a program from running:

    chalmers stop server1

Stopping a program will send a signal to the program. The signal can be set beforehand by:

    chalmers set server1 stopsignal=SIGTERM
 
Stop differs from off, off will not stop the program.
When off a program will not be started at system boot (or `start --all`)

When a program is *on* it will still start at system boot
even if you manually stop it before a reboot.
'''

from __future__ import unicode_literals, print_function

from argparse import RawDescriptionHelpFormatter
import logging
import sys

from clyent import print_colors

from chalmers import errors
from chalmers.utils import cli


log = logging.getLogger('chalmers.stop')

def main(args):

    programs = cli.select_programs(args, filter_paused=False, force=args.force)

    programs = cli.filter_programs(programs, lambda p: not p.is_running, 'Stopping', 'stopped')
    if not programs:
        return
    for prog in programs:
        if prog.is_running:
            print("Stopping program %-25s ... " % prog.name[:25], end=''); sys.stdout.flush()
            try:
                prog.stop(args.force)
            except errors.StateError as err:
                log.error(err.message)
            except errors.ConnectionError as err:
                print_colors("[  {=ERROR!c:red}  ] %s (use --force to force stop the program)" % err.message)
            else:
                print_colors("[  {=OK!c:green}  ]")
        else:
            print_colors("Program is already stopped: %-25s " % prog.name[:25], "[{=STOPPED!c:yello} ]")

def pause_main(args):

    programs = cli.select_programs(args, filter_paused=False)
    programs = cli.filter_programs(programs, lambda p: p.is_paused, 'Pausing', 'paused')
    if not programs:
        return

    for prog in programs:
        log.info("Pausing program %s" % (prog.name))
        if prog.is_running:
            log.warn("%s is running and will not restart on system reboot" % (prog.name))

        prog.state.update(paused=True)

def unpause_main(args):

    programs = cli.select_programs(args, filter_paused=False)
    programs = cli.filter_programs(programs, lambda p: not p.is_paused, 'Unpausing', 'unpaused')

    if not programs:
        return

    for prog in programs:
        log.info("Unpausing program %s" % (prog.name))
        prog.state.update(paused=False)
        if not prog.is_running:
            log.warning("%s is not running and will start on next system boot" % (prog.name))


def add_parser(subparsers):
    parser = subparsers.add_parser('stop',
                                   help='Stop running a command',
                                   description=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter)

    cli.add_selection_group(parser)
    parser.add_argument('--force', action='store_true',
                        help='Force kill a program (stopsignal will be ignored)'
                        )

    parser.set_defaults(main=main)

    parser = subparsers.add_parser('off',
                                   help='Don\'t run a program on system boot or `chalmers start --all`',
                                   description=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter)

    cli.add_selection_group(parser)

    parser.set_defaults(main=pause_main)

    parser = subparsers.add_parser('on',
                                   help='Run a program on system boot or `chalmers start --all`',
                                   description=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter)

    cli.add_selection_group(parser)

    parser.set_defaults(main=unpause_main)



"""

"""
import logging
import os
import signal
import sys

from chalmers import errors
from chalmers.utils.daemonize import daemonize

from .base import ProgramBase
import pwd


log = logging.getLogger(__name__)


def stop_process(signum, frame):
    """
    Signal handler to raise StopProcess exception
    """
    log.debug("Process received signal %s" % signum)
    raise errors.StopProcess()

class PosixProgram(ProgramBase):
    """
    Program that implements ProgramBase's abstract methods for posix platforms
    """

    @property
    def is_running(self):
        "Check For the existence of a pid"
        pid = self.state.get('pid')
        if not pid:
            return False
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def start_as_service(self):
        """
        Run this program in a new background process

        posix only
        """

        daemonize(self.start_sync)

    def sighup_handler(self, signum, frame):
        if self._ignore_sighup:
            log.warn("Program %s received signal %s ignoring" % (self.name, signum))
        else:
            log.error("Program %s received signal %s exiting" % (self.name, signum))
            sys.exit(-1)

    def handle_signals(self):
        self._ignore_sighup = False
        signal.signal(signal.SIGHUP, self.sighup_handler)

    def clear_socket(self):
        'Remove socket file'
        if os.path.exists(self.addr):
            log.debug("Removing socket file %s" % self.addr)
            try:
                os.unlink(self.addr)
            except OSError as err:
                log.warn("Error: %s" % err)


    def stop(self, force=False):
        try:
            ProgramBase.stop(self, force=force)
        finally:
            self.clear_socket()


    def dispatch_bg(self):
        log.warn("Moving process to the background")

        if self.pipe_output:
            print("Output is written to '%s'" % self.data['stdout'])
            self._echo.stop()

        log.warn("Ignoring sighup (%i)" % signal.SIGHUP)

        self._ignore_sighup = True

        si = file('/dev/null', 'r')
        so = file('/dev/null', 'a+')
        se = file('/dev/null', 'a+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())


    def preexec_fn(self):
        # IGNORE SIGHUP
        # This the chalmers managing process will catch the SIGHUP
        # and kill the child
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        new_mask = self.data.get('umask')
        user = self.data.get('user')

        if new_mask:
            os.umask(new_mask)

        if user:
            if isinstance(user, int):
                user = pwd.getpwuid(user)
            else:
                user = pwd.getpwnam(user)

            os.setegid(user.pw_gid)
            os.seteuid(user.pw_uid)


    def _send_signal(self, pid, sig):
        os.kill(pid, sig)





import os
from os.path import join
import tempfile
import unittest

from chalmers import config

if os.name == 'posix':
    from chalmers.program.posix import PosixProgram

@unittest.skipUnless(os.name == 'posix', 'Theses tests only run on posix systems')
class TestPosix(unittest.TestCase):

    def setUp(self):

        self.root_config = join(tempfile.gettempdir(), 'chalmers_tests')
        config.set_relative_dirs(self.root_config)
        unittest.TestCase.setUp(self)

    def test_init(self):

        p = PosixProgram('name', load=False)
        self.addCleanup(p.delete)
        self.assertEqual(p.name, 'name')

    def test_create(self):

        p = PosixProgram.create('name', {})
        self.addCleanup(p.delete)
        expected_keys = {u'stdout', u'redirect_stderr', u'stopwaitsecs', u'startsecs',
                         u'stopsignal', u'name', u'log_dir', u'startretries',
                         u'daemon_log', u'exitcodes'}
        self.assertEqual(set(p.data.keys()), expected_keys)
        self.assertEqual(p.state, {})





if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Chalmers main script

This script loads all of the chalmers commands
'''
from __future__ import print_function, unicode_literals

from argparse import ArgumentParser
import logging

from clyent import add_default_arguments, add_subparser_modules, run_command
from clyent.logs import setup_logging

from chalmers import __version__ as version
from chalmers import config
import chalmers.commands


logger = logging.getLogger('chalmers')

def main(args=None, exit=True):

    parser = ArgumentParser(description=__doc__)

    add_default_arguments(parser, version)
    add_subparser_modules(parser, chalmers.commands)

    args = parser.parse_args(args)
    logfile = config.main_logfile()
    setup_logging(logger, args.log_level, use_color=args.color,
                  show_tb=args.show_traceback, logfile=logfile)

    if not hasattr(args, 'main'):
        parser.error('too few arguments')

    run_command(args, exit=exit)

if __name__ == "__main__":
    main()
